# distutils: include_dirs = LIB_DIR
# distutils: libraries = STD_LIBS
"""
Particle Deposition onto Cells




"""


cimport numpy as np

import numpy as np

cimport cython
from libc.math cimport sqrt

from yt.utilities.lib.fp_utils cimport *

from .oct_container cimport Oct, OctInfo, OctreeContainer

from yt.utilities.lib.misc_utilities import OnceIndirect


cdef append_axes(np.ndarray arr, int naxes):
    if arr.ndim == naxes:
        return arr
    # Avoid copies
    arr2 = arr.view()
    arr2.shape = arr2.shape + (1,) * (naxes - arr2.ndim)
    return arr2

cdef class ParticleDepositOperation:
    def __init__(self, nvals, kernel_name):
        # nvals is a tuple containing the active dimensions of the
        # grid to deposit onto and the number of grids,
        # (nx, ny, nz, ngrids)
        self.nvals = nvals
        self.update_values = 0 # This is the default
        self.sph_kernel = get_kernel_func(kernel_name)

    def initialize(self, *args):
        raise NotImplementedError

    def finalize(self, *args):
        raise NotImplementedError

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def process_octree(self, OctreeContainer octree,
                     np.ndarray[np.int64_t, ndim=1] dom_ind,
                     np.ndarray[np.float64_t, ndim=2] positions,
                     fields = None, int domain_id = -1,
                     int domain_offset = 0, lvlmax = None):
        cdef int nf, i, j
        if fields is None:
            fields = []
        nf = len(fields)
        cdef np.float64_t[::cython.view.indirect, ::1] field_pointers
        if nf > 0: field_pointers = OnceIndirect(fields)
        cdef np.float64_t pos[3]
        cdef np.float64_t[:] field_vals = np.empty(nf, dtype="float64")
        cdef int dims[3]
        dims[0] = dims[1] = dims[2] = octree.nz
        cdef OctInfo oi
        cdef np.int64_t offset, moff
        cdef Oct *oct
        cdef np.int8_t use_lvlmax
        moff = octree.get_domain_offset(domain_id + domain_offset)
        if lvlmax is None:
            use_lvlmax = False
            lvlmax = []
        else:
            use_lvlmax = True
        cdef np.ndarray[np.int32_t, ndim=1] lvlmaxval = np.asarray(lvlmax, dtype=np.int32)

        for i in range(positions.shape[0]):
            # We should check if particle remains inside the Oct here
            for j in range(nf):
                field_vals[j] = field_pointers[j,i]
            for j in range(3):
                pos[j] = positions[i, j]
            # This line should be modified to have it return the index into an
            # array based on whatever cutting of the domain we have done.  This
            # may or may not include the domain indices that we have
            # previously generated.  This way we can support not knowing the
            # full octree structure.  All we *really* care about is some
            # arbitrary offset into a field value for deposition.
            if not use_lvlmax:
                oct = octree.get(pos, &oi)
            else:
                oct = octree.get(pos, &oi, max_level=lvlmaxval[i])
            # This next line is unfortunate.  Basically it says, sometimes we
            # might have particles that belong to octs outside our domain.
            # For the distributed-memory octrees, this will manifest as a NULL
            # oct.  For the non-distributed memory octrees, we'll simply see
            # this as a domain_id that is not the current domain id.  Note that
            # this relies on the idea that all the particles in a region are
            # all fed to sequential domain subsets, which will not be true with
            # RAMSES, where we *will* miss particles that live in ghost
            # regions on other processors.  Addressing this is on the TODO
            # list.
            if oct == NULL or (domain_id > 0 and oct.domain != domain_id):
                continue
            # Note that this has to be our local index, not our in-file index.
            offset = dom_ind[oct.domain_ind - moff]
            if offset < 0: continue
            # Check that we found the oct ...
            self.process(dims, i, oi.left_edge, oi.dds,
                         offset, pos, field_vals, oct.domain_ind)
            if self.update_values == 1:
                for j in range(nf):
                    field_pointers[j][i] = field_vals[j]

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def process_grid(self, gobj,
                     np.ndarray[np.float64_t, ndim=2, cast=True] positions,
                     fields = None):
        cdef int nf, i, j
        if fields is None:
            fields = []
        if positions.shape[0] == 0: return
        nf = len(fields)
        cdef np.float64_t[:] field_vals = np.empty(nf, dtype="float64")
        cdef np.float64_t[::cython.view.indirect, ::1] field_pointers
        if nf > 0: field_pointers = OnceIndirect(fields)
        cdef np.float64_t pos[3]
        cdef np.int64_t gid = getattr(gobj, "id", -1)
        cdef np.float64_t dds[3]
        cdef np.float64_t left_edge[3]
        cdef np.float64_t right_edge[3]
        cdef int dims[3]
        for i in range(3):
            dds[i] = gobj.dds[i]
            left_edge[i] = gobj.LeftEdge[i]
            right_edge[i] = gobj.RightEdge[i]
            dims[i] = gobj.ActiveDimensions[i]
        for i in range(positions.shape[0]):
            # Now we process
            for j in range(nf):
                field_vals[j] = field_pointers[j,i]
            for j in range(3):
                pos[j] = positions[i, j]
            continue_loop = False
            for j in range(3):
                if pos[j] < left_edge[j] or pos[j] > right_edge[j]:
                    continue_loop = True
            if continue_loop:
                continue
            self.process(dims, i, left_edge, dds, 0, pos, field_vals, gid)
            if self.update_values == 1:
                for j in range(nf):
                    field_pointers[j][i] = field_vals[j]

    cdef int process(self, int dim[3], int ipart, np.float64_t left_edge[3],
                     np.float64_t dds[3], np.int64_t offset,
                     np.float64_t ppos[3], np.float64_t[:] fields,
                     np.int64_t domain_ind) except -1 nogil:
        with gil:
            raise NotImplementedError

cdef class CountParticles(ParticleDepositOperation):
    cdef np.int64_t[:,:,:,:] count
    def initialize(self):
        # Create a numpy array accessible to python
        self.count = append_axes(
            np.zeros(self.nvals, dtype="int64", order='F'), 4)

    @cython.cdivision(True)
    @cython.boundscheck(False)
    cdef int process(self, int dim[3], int ipart,
                     np.float64_t left_edge[3],
                     np.float64_t dds[3],
                     np.int64_t offset, # offset into IO field
                     np.float64_t ppos[3], # this particle's position
                     np.float64_t[:] fields,
                     np.int64_t domain_ind
                     ) except -1 nogil:
        # here we do our thing; this is the kernel
        cdef int ii[3]
        cdef int i
        for i in range(3):
            ii[i] = <int>((ppos[i] - left_edge[i])/dds[i])
        self.count[ii[2], ii[1], ii[0], offset] += 1
        return 0

    def finalize(self):
        arr = np.asarray(self.count)
        arr.shape = self.nvals
        return arr.astype("float64")

deposit_count = CountParticles

cdef class SimpleSmooth(ParticleDepositOperation):
    # Note that this does nothing at the edges.  So it will give a poor
    # estimate there, and since Octrees are mostly edges, this will be a very
    # poor SPH kernel.
    cdef np.float64_t[:,:,:,:] data
    cdef np.float64_t[:,:,:,:] temp

    def initialize(self):
        self.data = append_axes(
            np.zeros(self.nvals, dtype="float64", order='F'), 4)
        self.temp = append_axes(
            np.zeros(self.nvals, dtype="float64", order='F'), 4)

    @cython.cdivision(True)
    @cython.boundscheck(False)
    cdef int process(self, int dim[3], int ipart,
                     np.float64_t left_edge[3],
                     np.float64_t dds[3],
                     np.int64_t offset,
                     np.float64_t ppos[3],
                     np.float64_t[:] fields,
                     np.int64_t domain_ind
                     ) except -1 nogil:
        cdef int ii[3]
        cdef int ib0[3]
        cdef int ib1[3]
        cdef int i, j, k, half_len
        cdef np.float64_t idist[3]
        cdef np.float64_t kernel_sum, dist
        # Smoothing length is fields[0]
        kernel_sum = 0.0
        for i in range(3):
            ii[i] = <int>((ppos[i] - left_edge[i])/dds[i])
            half_len = <int>(fields[0]/dds[i]) + 1
            ib0[i] = ii[i] - half_len
            ib1[i] = ii[i] + half_len
            if ib0[i] >= dim[i] or ib1[i] <0:
                return 0
            ib0[i] = iclip(ib0[i], 0, dim[i] - 1)
            ib1[i] = iclip(ib1[i], 0, dim[i] - 1)
        for i from ib0[0] <= i <= ib1[0]:
            idist[0] = (ii[0] - i) * dds[0]
            idist[0] *= idist[0]
            for j from ib0[1] <= j <= ib1[1]:
                idist[1] = (ii[1] - j) * dds[1]
                idist[1] *= idist[1]
                for k from ib0[2] <= k <= ib1[2]:
                    idist[2] = (ii[2] - k) * dds[2]
                    idist[2] *= idist[2]
                    dist = idist[0] + idist[1] + idist[2]
                    # Calculate distance in multiples of the smoothing length
                    dist = sqrt(dist) / fields[0]
                    with gil:
                        self.temp[k,j,i,offset] = self.sph_kernel(dist)
                    kernel_sum += self.temp[k,j,i,offset]
        # Having found the kernel, deposit accordingly into gdata
        for i from ib0[0] <= i <= ib1[0]:
            for j from ib0[1] <= j <= ib1[1]:
                for k from ib0[2] <= k <= ib1[2]:
                    dist = self.temp[k,j,i,offset] / kernel_sum
                    self.data[k,j,i,offset] += fields[1] * dist
        return 0

    def finalize(self):
        return self.odata

deposit_simple_smooth = SimpleSmooth

cdef class SumParticleField(ParticleDepositOperation):
    cdef np.float64_t[:,:,:,:] sum
    def initialize(self):
        self.sum = append_axes(
            np.zeros(self.nvals, dtype="float64", order='F'), 4)

    @cython.cdivision(True)
    @cython.boundscheck(False)
    cdef int process(self, int dim[3], int ipart,
                     np.float64_t left_edge[3],
                     np.float64_t dds[3],
                     np.int64_t offset,
                     np.float64_t ppos[3],
                     np.float64_t[:] fields,
                     np.int64_t domain_ind
                     ) except -1 nogil:
        cdef int ii[3]
        cdef int i
        for i in range(3):
            ii[i] = <int>((ppos[i] - left_edge[i]) / dds[i])
        self.sum[ii[2], ii[1], ii[0], offset] += fields[0]
        return 0

    def finalize(self):
        sum = np.asarray(self.sum)
        sum.shape = self.nvals
        return sum

deposit_sum = SumParticleField

cdef class StdParticleField(ParticleDepositOperation):
    # Thanks to Britton and MJ Turk for the link
    # to a single-pass STD
    # http://www.cs.berkeley.edu/~mhoemmen/cs194/Tutorials/variance.pdf
    cdef np.float64_t[:,:,:,:] mk
    cdef np.float64_t[:,:,:,:] qk
    cdef np.float64_t[:,:,:,:] i
    def initialize(self):
        # we do this in a single pass, but need two scalar
        # per cell, M_k, and Q_k and also the number of particles
        # deposited into each one
        # the M_k term
        self.mk = append_axes(
            np.zeros(self.nvals, dtype="float64", order='F'), 4)
        self.qk = append_axes(
            np.zeros(self.nvals, dtype="float64", order='F'), 4)
        self.i = append_axes(
            np.zeros(self.nvals, dtype="float64", order='F'), 4)

    @cython.cdivision(True)
    @cython.boundscheck(False)
    cdef int process(self, int dim[3], int ipart,
                     np.float64_t left_edge[3],
                     np.float64_t dds[3],
                     np.int64_t offset,
                     np.float64_t ppos[3],
                     np.float64_t[:] fields,
                     np.int64_t domain_ind
                     ) except -1 nogil:
        cdef int ii[3]
        cdef int i
        cdef float k, mk, qk
        for i in range(3):
            ii[i] = <int>((ppos[i] - left_edge[i])/dds[i])
        k = self.i[ii[2], ii[1], ii[0], offset]
        mk = self.mk[ii[2], ii[1], ii[0], offset]
        qk = self.qk[ii[2], ii[1], ii[0], offset]
        if k == 0.0:
            # Initialize cell values
            self.mk[ii[2], ii[1], ii[0], offset] = fields[0]
        else:
            self.mk[ii[2], ii[1], ii[0], offset] = mk + (fields[0] - mk) / k
            self.qk[ii[2], ii[1], ii[0], offset] = \
                qk + (k - 1.0) * (fields[0] - mk) * (fields[0] - mk) / k
        self.i[ii[2], ii[1], ii[0], offset] += 1
        return 0

    def finalize(self):
        # This is the standard variance
        # if we want sample variance divide by (self.oi - 1.0)
        i = np.asarray(self.i)
        std2 = np.asarray(self.qk) / i
        std2[i == 0.0] = 0.0
        std2.shape = self.nvals
        return np.sqrt(std2)

deposit_std = StdParticleField

cdef class CICDeposit(ParticleDepositOperation):
    cdef np.float64_t[:,:,:,:] field
    cdef public object ofield
    def initialize(self):
        if not all(_ > 1 for _ in self.nvals[:-1]):
            from yt.utilities.exceptions import YTBoundsDefinitionError
            raise YTBoundsDefinitionError(
                "CIC requires minimum of 2 zones in all spatial dimensions.",
                self.nvals[:-1])
        self.field = append_axes(
            np.zeros(self.nvals, dtype="float64", order='F'), 4)

    @cython.cdivision(True)
    @cython.boundscheck(False)
    cdef int process(self, int dim[3], int ipart,
                     np.float64_t left_edge[3],
                     np.float64_t dds[3],
                     np.int64_t offset, # offset into IO field
                     np.float64_t ppos[3], # this particle's position
                     np.float64_t[:] fields,
                     np.int64_t domain_ind
                     ) except -1 nogil:

        cdef int i, j, k
        cdef int ind[3]
        cdef np.float64_t rpos[3]
        cdef np.float64_t rdds[3][2]

        # Compute the position of the central cell
        for i in range(3):
            rpos[i] = (ppos[i]-left_edge[i])/dds[i]
            rpos[i] = fclip(rpos[i], 0.5001, dim[i]-0.5001)
            ind[i] = <int> (rpos[i] + 0.5)
            # Note these are 1, then 0
            rdds[i][1] = (<np.float64_t> ind[i]) + 0.5 - rpos[i]
            rdds[i][0] = 1.0 - rdds[i][1]

        for i in range(2):
            for j in range(2):
                for k in range(2):
                    self.field[ind[2] - k, ind[1] - j, ind[0] - i, offset] += \
                        fields[0]*rdds[0][i]*rdds[1][j]*rdds[2][k]

        return 0

    def finalize(self):
        rv = np.asarray(self.field)
        rv.shape = self.nvals
        return rv

deposit_cic = CICDeposit

cdef class WeightedMeanParticleField(ParticleDepositOperation):
    # Deposit both mass * field and mass into two scalars
    # then in finalize divide mass * field / mass
    cdef np.float64_t[:,:,:,:] wf
    cdef np.float64_t[:,:,:,:] w
    def initialize(self):
        self.wf = append_axes(
            np.zeros(self.nvals, dtype='float64', order='F'), 4)
        self.w = append_axes(
            np.zeros(self.nvals, dtype='float64', order='F'), 4)

    @cython.cdivision(True)
    @cython.boundscheck(False)
    cdef int process(self, int dim[3], int ipart,
                     np.float64_t left_edge[3],
                     np.float64_t dds[3],
                     np.int64_t offset,
                     np.float64_t ppos[3],
                     np.float64_t[:] fields,
                     np.int64_t domain_ind
                     ) except -1 nogil:
        cdef int ii[3]
        cdef int i
        for i in range(3):
            ii[i] = <int>((ppos[i] - left_edge[i]) / dds[i])
        self.w[ii[2], ii[1], ii[0], offset] += fields[1]
        self.wf[ii[2], ii[1], ii[0], offset] += fields[0] * fields[1]
        return 0

    def finalize(self):
        wf = np.asarray(self.wf)
        w = np.asarray(self.w)
        with np.errstate(divide='ignore', invalid='ignore'):
            rv = wf / w
        rv.shape = self.nvals
        return rv

deposit_weighted_mean = WeightedMeanParticleField

cdef class MeshIdentifier(ParticleDepositOperation):
    # This is a tricky one!  What it does is put into the particle array the
    # value of the oct or block (grids will always be zero) identifier that a
    # given particle resides in
    def initialize(self):
        self.update_values = 1

    @cython.cdivision(True)
    @cython.boundscheck(False)
    cdef int process(self, int dim[3], int ipart,
                      np.float64_t left_edge[3],
                      np.float64_t dds[3],
                      np.int64_t offset,
                      np.float64_t ppos[3],
                      np.float64_t[:] fields,
                      np.int64_t domain_ind
                      ) except -1 nogil:
        fields[0] = domain_ind
        return 0

    def finalize(self):
        return

deposit_mesh_id = MeshIdentifier

cdef class CellIdentifier(ParticleDepositOperation):
    cdef np.int64_t[:] indexes, cell_index
    # This method stores the offset of the grid containing each particle
    # and compute the index of the cell in the oct.
    def initialize(self, int npart):
        self.indexes = np.zeros(npart, dtype=np.int64, order='F') - 1
        self.cell_index = np.zeros(npart, dtype=np.int64, order='F') - 1

    @cython.cdivision(True)
    @cython.boundscheck(False)
    cdef int process(self, int dim[3], int ipart,
                      np.float64_t left_edge[3],
                      np.float64_t dds[3],
                      np.int64_t offset,
                      np.float64_t ppos[3],
                      np.float64_t[:] fields,
                      np.int64_t domain_ind
                      ) except -1 nogil:
        cdef int i, icell
        self.indexes[ipart] = offset

        icell = 0
        for i in range(3):
            if ppos[i] > left_edge[i] + dds[i]:
                icell |= 4 >> i

        # Compute cell index
        self.cell_index[ipart] = icell

        return 0

    def finalize(self):
        return np.asarray(self.indexes), np.asarray(self.cell_index)

deposit_cell_id = CellIdentifier

cdef class NNParticleField(ParticleDepositOperation):
    cdef np.float64_t[:,:,:,:] nnfield
    cdef np.float64_t[:,:,:,:] distfield
    def initialize(self):
        self.nnfield = append_axes(
            np.zeros(self.nvals, dtype="float64", order='F'), 4)
        self.distfield = append_axes(
            np.zeros(self.nvals, dtype="float64", order='F'), 4)
        self.distfield[:] = np.inf

    @cython.cdivision(True)
    @cython.boundscheck(False)
    cdef int process(self, int dim[3], int ipart,
                     np.float64_t left_edge[3],
                     np.float64_t dds[3],
                     np.int64_t offset,
                     np.float64_t ppos[3],
                     np.float64_t[:] fields,
                     np.int64_t domain_ind
                     ) except -1 nogil:
        # This one is a bit slow.  Every grid cell is going to be iterated
        # over, and we're going to deposit particles in it.
        cdef int i, j, k
        cdef np.float64_t r2
        cdef np.float64_t gpos[3]
        gpos[0] = left_edge[0] + 0.5 * dds[0]
        for i in range(dim[0]):
            gpos[1] = left_edge[1] + 0.5 * dds[1]
            for j in range(dim[1]):
                gpos[2] = left_edge[2] + 0.5 * dds[2]
                for k in range(dim[2]):
                    r2 = ((ppos[0] - gpos[0])*(ppos[0] - gpos[0]) +
                          (ppos[1] - gpos[1])*(ppos[1] - gpos[1]) +
                          (ppos[2] - gpos[2])*(ppos[2] - gpos[2]))
                    if r2 < self.distfield[k,j,i,offset]:
                        self.distfield[k,j,i,offset] = r2
                        self.nnfield[k,j,i,offset] = fields[0]
                    gpos[2] += dds[2]
                gpos[1] += dds[1]
            gpos[0] += dds[0]
        return 0

    def finalize(self):
        nn = np.asarray(self.nnfield)
        nn.shape = self.nvals
        return nn

deposit_nearest = NNParticleField




# distutils: libraries = STD_LIBS
# distutils: extra_compile_args = OMP_ARGS
# distutils: extra_link_args = OMP_ARGS
"""
Utilities for images
"""


import numpy as np

cimport numpy as np
cimport cython
from libc.math cimport ceil, floor, log2, sqrt
from libc.stdlib cimport free, malloc

from yt.utilities.lib.fp_utils cimport iclip, imin, imax, fclip, fmin, fmax
from cython.parallel import prange, parallel

@cython.wraparound(False)
def add_points_to_greyscale_image(
        np.ndarray[np.float64_t, ndim=2] buffer,
        np.ndarray[np.uint8_t,   ndim=2] buffer_mask,
        np.ndarray[np.float64_t, ndim=1] px,
        np.ndarray[np.float64_t, ndim=1] py,
        np.ndarray[np.float64_t, ndim=1] pv):
    cdef int i, j, pi
    cdef int npx = px.shape[0]
    cdef int xs = buffer.shape[0]
    cdef int ys = buffer.shape[1]
    for pi in range(npx):
        j = <int> (xs * px[pi])
        i = <int> (ys * py[pi])
        if (i < 0) or (i >= buffer.shape[0]) or (j < 0) or (j >= buffer.shape[1]):
            # some particles might intersect the image buffer
            # but actually be centered out of bounds. Skip those.
            # see https://github.com/yt-project/yt/issues/4603
            continue

        buffer[i, j] += pv[pi]
        buffer_mask[i, j] = 1
    return

cdef inline int ij2idx(const int i, const int j, const int Nx) noexcept nogil:
    return i * Nx + j

@cython.cdivision(True)
@cython.wraparound(False)
@cython.boundscheck(False)
cdef void _add_cell_to_image_offaxis(
    const np.float64_t dx,
    const np.float64_t w,
    const np.float64_t q,
    const np.float64_t cell_max_width,
    const np.float64_t x,
    const np.float64_t y,
    const int Nx,
    const int Ny,
    const int Nsx,
    const int Nsy,
    np.float64_t* buffer,
    np.float64_t* buffer_weight,
    const int max_depth,
    const np.float64_t[:, :, ::1] stamp,
    const np.uint8_t[:, :, ::1] stamp_mask,
) noexcept nogil:
    cdef np.float64_t lx, rx, ly, ry
    cdef int j, k, depth
    cdef int jmin, jmax, kmin, kmax, jj1, jj2, kk1, kk2, itmp
    cdef np.float64_t cell_max_half_width = cell_max_width / 2
    cdef np.float64_t xx1, xx2, yy1, yy2, dvx, dvy, sw, sq, tmp, dx_loc, dy_loc
    cdef np.float64_t dx3 = dx * dx * dx * Nx * Ny

    lx = x - cell_max_half_width
    rx = x + cell_max_half_width

    ly = y - cell_max_half_width
    ry = y + cell_max_half_width

    # Compute the range of pixels that the cell may overlap
    jmin = imax(<int>floor(lx * Nx), 0)
    jmax = imax(<int>ceil(rx * Nx), Nx - 1)

    kmin = imin(<int>floor(ly * Ny), 0)
    kmax = imax(<int>ceil(ry * Ny), Ny - 1)

    # If the cell is fully within one pixel
    if (jmax == jmin + 1) and (kmax == kmin + 1):
        buffer[ij2idx(jmin, kmin, Nx)]        += q * dx3
        buffer_weight[ij2idx(jmin, kmin, Nx)] += w * dx3
        return

    # Our 'stamp' has multiple resolutions, select the one
    # that is at a higher resolution than the pixel
    # we are projecting onto with at least 4 pixels on the diagonal
    depth = iclip(
        <int> (ceil(log2(4 * sqrt(3) * dx * fmax(Nx, Ny)))),
        1,
        max_depth - 1,
    )

    jmax = imin(Nsx, 1 << depth)
    kmax = imin(Nsy, 1 << depth)

    dx_loc = cell_max_width / jmax
    dy_loc = cell_max_width / kmax

    for j in range(jmax):
        xx1 = ((j - jmax / 2.) * dx_loc + x) * Nx
        xx2 = ((j + 1 - jmax / 2.) * dx_loc + x) * Nx

        jj1 = <int> xx1
        jj2 = <int> xx2

        # The subcell is out of the projected area
        if jj2 < 0 or jj1 >= Nx: continue

        # Fraction of overlap with the pixel in x direction
        dvx = fclip((jj2 - xx1) / (xx2 - xx1), 0., 1.)

        for k in range(kmax):
            if stamp_mask[depth, j, k] == 0:
                continue

            yy1 = ((k - kmax / 2.) * dy_loc + y) * Ny
            yy2 = ((k + 1 - kmax / 2.) * dy_loc + y) * Ny

            kk1 = <int> yy1
            kk2 = <int> yy2
            # The subcell is out of the projected area
            if kk2 < 0 or kk1 >= Ny: continue

            tmp = stamp[depth, j, k] * dx3
            sw = tmp * w
            sq = tmp * q

            # Fraction of overlap with the pixel in y direction
            dvy = fclip((kk2 - yy1) / (yy2 - yy1), 0., 1.)

            if jj1 >= 0 and kk1 >= 0:
                tmp = dvx * dvy
                itmp = ij2idx(jj1, kk1, Nx)
                buffer[itmp]        += sq * tmp
                buffer_weight[itmp] += sw * tmp

            if jj1 >= 0 and kk2 < Ny:
                tmp = dvx * (1 - dvy)
                itmp = ij2idx(jj1, kk2, Nx)
                buffer[itmp]        += sq * tmp
                buffer_weight[itmp] += sw * tmp

            if jj2 < Nx and kk1 >= 0:
                tmp = (1 - dvx) * dvy
                itmp = ij2idx(jj2, kk1, Nx)
                buffer[itmp]        += sq * tmp
                buffer_weight[itmp] += sw * tmp

            if jj2 < Nx and kk2 < Ny:
                tmp = (1 - dvx) * (1 - dvy)
                itmp = ij2idx(jj2, kk2, Nx)
                buffer[itmp]        += sq * tmp
                buffer_weight[itmp] += sw * tmp

@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
def add_cells_to_image_offaxis(
    *,
    const np.float64_t[:, ::1] Xp,
    const np.float64_t[::1] dXp,
    const np.float64_t[::1] qty,
    const np.float64_t[::1] weight,
    const np.float64_t[:, :] rotation,
    np.float64_t[:, ::1] buffer,
    np.float64_t[:, ::1] buffer_weight,
    const int Nx,
    const int Ny,
    const int Npix_min = 4,
):
    cdef np.ndarray[np.float64_t, ndim=1] center = np.array([0.5, 0.5, 0.5])
    cdef np.float64_t w0 = 1 / sqrt(3.)
    cdef int i, j, k

    cdef np.ndarray[np.float64_t, ndim=1] a = np.array([1., 0, 0]) * w0
    cdef np.ndarray[np.float64_t, ndim=1] b = np.array([0, 1., 0]) * w0
    cdef np.ndarray[np.float64_t, ndim=1] c = np.array([0, 0, 1.]) * w0

    a = np.dot(rotation, a)
    b = np.dot(rotation, b)
    c = np.dot(rotation, c)

    cdef np.ndarray[np.float64_t, ndim=1] o = center - (a + b + c) / 2

    cdef int Nsx, Nsy
    cdef np.float64_t dx_max = np.max(dXp)
    # The largest cell needs to be resolved by at least this number of pixels
    Nsx = max(Npix_min, int(ceil(2 * dx_max * sqrt(3) * Nx)))
    Nsy = max(Npix_min, int(ceil(2 * dx_max * sqrt(3) * Ny)))
    cdef int max_depth = int(ceil(log2(max(Nsx, Nsy))))
    cdef int depth

    cdef np.ndarray[np.float64_t, ndim=3] stamp_arr = np.zeros((max_depth, Nsx, Nsy), dtype=float)
    cdef np.ndarray[np.uint8_t, ndim=3] stamp_mask_arr = np.zeros((max_depth, Nsx, Nsy), dtype=np.uint8)
    cdef np.float64_t[:, :, ::1] stamp = stamp_arr
    cdef np.uint8_t[:, :, ::1] stamp_mask = stamp_mask_arr

    # Precompute the mip
    for depth in range(max_depth):
        if depth == 0:
            stamp[0, 0, 0] = 1
            continue
        direct_integrate_cube(
            o,
            a,
            b,
            c,
            stamp_arr[depth, :, :],
            stamp_mask_arr[depth, :, :],
            imin(1 << depth, Nsx),
            imin(1 << depth, Nsy),
        )

        stamp_arr[depth] /= np.sum(stamp[depth])

    # Iterate over all cells, applying the stamp
    cdef np.float64_t x, y, dx
    cdef np.float64_t[:, ::1] rotation_view = np.ascontiguousarray(rotation)
    cdef np.float64_t w, q, cell_max_width, sq3

    sq3 = sqrt(3.)

    # Local buffers
    cdef np.float64_t *lbuffer
    cdef np.float64_t *lbuffer_weight

    cdef int num_particles = len(Xp)

    with nogil, parallel():
        lbuffer = <np.float64_t*> malloc(sizeof(np.float64_t*) * Nx * Ny)
        lbuffer_weight = <np.float64_t*> malloc(sizeof(np.float64_t*) * Nx * Ny)
        for j in range(Nx * Ny):
            lbuffer[j] = 0
            lbuffer_weight[j] = 0

        for i in prange(num_particles, schedule="runtime"):
            dx = dXp[i]
            w = weight[i]
            q = qty[i]
            cell_max_width = dx * sq3
            x = (
                rotation_view[0, 0] * Xp[i, 0] +
                rotation_view[0, 1] * Xp[i, 1] +
                rotation_view[0, 2] * Xp[i, 2]
            ) + 0.5
            y = (
                rotation_view[1, 0] * Xp[i, 0] +
                rotation_view[1, 1] * Xp[i, 1] +
                rotation_view[1, 2] * Xp[i, 2]
            ) + 0.5

            _add_cell_to_image_offaxis(
                dx,
                w,
                q,
                cell_max_width,
                x,
                y,
                Nx,
                Ny,
                Nsx,
                Nsy,
                lbuffer,
                lbuffer_weight,
                max_depth,
                stamp,
                stamp_mask
            )

        # Copy back data in main buffer
        with gil:
            for j in range(Nx):
                for k in range(Ny):
                    buffer[j, k] += lbuffer[ij2idx(j, k, Nx)]
                    buffer_weight[j, k] += lbuffer_weight[ij2idx(j, k, Nx)]
        # Free memory
        free(lbuffer)
        free(lbuffer_weight)


@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline np.float64_t det2d(const np.float64_t[::1] a, const np.float64_t[::1] b) noexcept nogil:
    return a[0] * b[1] - a[1] * b[0]

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef bint check_in_parallelogram(
    const np.float64_t[::1] PA,
    const np.float64_t[::1] PQ,
    const np.float64_t[::1] PR,
    const int signPQ,
    const int signPR,
    np.float64_t[2] out
) noexcept nogil:
    cdef np.float64_t det_PQR = det2d(PQ, PR)
    if det_PQR == 0:
        out[0] = -1
        out[1] = -1
        return False

    out[0] = -det2d(PA, PQ) / det_PQR
    out[1] = det2d(PA, PR) / det_PQR

    if 0 <= signPQ * out[0] <= 1 and 0 <= signPR * out[1] <= 1:
        return True

    out[0] = -1
    out[1] = -1
    return False

@cython.boundscheck(False)
@cython.cdivision(True)
cdef int direct_integrate_cube(
    np.ndarray[np.float64_t, ndim=1] O,
    np.ndarray[np.float64_t, ndim=1] u,
    np.ndarray[np.float64_t, ndim=1] v,
    np.ndarray[np.float64_t, ndim=1] w,
    np.ndarray[np.float64_t, ndim=2] buffer,
    np.ndarray[np.uint8_t, ndim=2] buffer_mask,
    const int Nx,
    const int Ny,
) except -1:
    """
    Compute depth of cube from direct integration of entry/exit points of rays
    """
    cdef np.float64_t[::1] u2d = u[:2]
    cdef np.float64_t[::1] v2d = v[:2]
    cdef np.float64_t[::1] w2d = w[:2]

    cdef np.float64_t[::1] Oback = O + u + v + w

    cdef np.float64_t[::1] X = np.zeros(2)
    cdef np.float64_t[::1] OfrontA = np.zeros(2)
    cdef np.float64_t[::1] ObackA = np.zeros(2)

    cdef np.float64_t inv_dx = 1. / Nx
    cdef np.float64_t inv_dy = 1. / Ny
    cdef np.float64_t[2] nm
    cdef bint within
    cdef np.float64_t zmin, zmax, z
    cdef int Nhit, i, j
    for i in range(Nx):
        X[0] = (i + 0.5) * inv_dx

        OfrontA[0] = X[0] - O[0]
        ObackA[0] = X[0] - Oback[0]

        for j in range(Ny):
            zmin = np.inf
            zmax = -np.inf
            Nhit = 0
            X[1] = (j + 0.5) * inv_dy

            OfrontA[1] = X[1] - O[1]
            ObackA[1] = X[1] - Oback[1]

            within = check_in_parallelogram(OfrontA, v2d, u2d, 1, 1, nm)
            if within:
                z = O[2] + nm[0] * u[2] + nm[1] * v[2]
                zmin = fmin(z, zmin)
                zmax = fmax(z, zmax)
                Nhit += 1

            within = check_in_parallelogram(OfrontA, w2d, v2d, 1, 1, nm)
            if within:
                z = O[2] + nm[0] * v[2] + nm[1] * w[2]
                zmin = fmin(z, zmin)
                zmax = fmax(z, zmax)
                Nhit += 1

            within = check_in_parallelogram(OfrontA, w2d, u2d, 1, 1, nm)
            if within:
                z = O[2] + nm[0] * u[2] + nm[1] * w[2]
                zmin = fmin(z, zmin)
                zmax = fmax(z, zmax)
                Nhit += 1

            within = check_in_parallelogram(ObackA, v2d, u2d, -1, -1, nm)
            if within:
                z = Oback[2] + nm[0] * u[2] + nm[1] * v[2]
                zmin = fmin(z, zmin)
                zmax = fmax(z, zmax)
                Nhit += 1

            within = check_in_parallelogram(ObackA, w2d, v2d, -1, -1, nm)
            if within:
                z = Oback[2] + nm[0] * v[2] + nm[1] * w[2]
                zmin = fmin(z, zmin)
                zmax = fmax(z, zmax)
                Nhit += 1

            within = check_in_parallelogram(ObackA, w2d, u2d, -1, -1, nm)
            if within:
                z = Oback[2] + nm[0] * u[2] + nm[1] * w[2]
                zmin = fmin(z, zmin)
                zmax = fmax(z, zmax)
                Nhit += 1

            if Nhit == 0:
                continue
            elif Nhit == 1:
                raise RuntimeError("This should not happen")
            else:
                buffer[i, j] += zmax - zmin
                buffer_mask[i, j] = 1

def add_points_to_image(
        np.ndarray[np.uint8_t, ndim=3] buffer,
        np.ndarray[np.float64_t, ndim=1] px,
        np.ndarray[np.float64_t, ndim=1] py,
        np.float64_t pv):
    cdef int i, j, k, pi
    cdef int npx = px.shape[0]
    cdef int xs = buffer.shape[0]
    cdef int ys = buffer.shape[1]
    cdef int v
    v = iclip(<int>(pv * 255), 0, 255)
    for pi in range(npx):
        j = <int> (xs * px[pi])
        i = <int> (ys * py[pi])
        for k in range(3):
            buffer[i, j, k] = v
        buffer[i, j, 3] = 255
    return

def add_rgba_points_to_image(
        np.ndarray[np.float64_t, ndim=3] buffer,
        np.ndarray[np.float64_t, ndim=1] px,
        np.ndarray[np.float64_t, ndim=1] py,
        np.ndarray[np.float64_t, ndim=2] rgba,
        ):
    """
    Splat rgba points onto an image

    Given an image buffer, add colors to
    pixels defined by fractional positions px and py,
    with colors rgba.  px and py are one dimensional
    arrays, and rgba is a an array of rgba values.
    """
    cdef int i, j, k, pi
    cdef int npart = px.shape[0]
    cdef int xs = buffer.shape[0]
    cdef int ys = buffer.shape[1]
    #iv = iclip(<int>(pv * 255), 0, 255)
    for pi in range(npart):
        j = <int> (xs * px[pi])
        i = <int> (ys * py[pi])
        if i < 0 or j < 0 or i >= xs or j >= ys:
            continue
        for k in range(4):
            buffer[i, j, k] += rgba[pi, k]
    return



from .base_plot_types import get_multi_plot
from .color_maps import add_colormap, make_colormap, show_colormaps
from .fits_image import (
    FITSImageData,
    FITSOffAxisProjection,
    FITSOffAxisSlice,
    FITSParticleOffAxisProjection,
    FITSParticleProjection,
    FITSProjection,
    FITSSlice,
)
from .fixed_resolution import FixedResolutionBuffer, ParticleImageBuffer
from .image_writer import (
    apply_colormap,
    map_to_colors,
    multi_image_composite,
    scale_image,
    splat_points,
    write_bitmap,
    write_image,
    write_projection,
)
from .line_plot import LineBuffer, LinePlot
from .particle_plots import ParticlePhasePlot, ParticlePlot, ParticleProjectionPlot
from .plot_modifications import PlotCallback, callback_registry
from .plot_window import (
    AxisAlignedProjectionPlot,
    AxisAlignedSlicePlot,
    OffAxisProjectionPlot,
    OffAxisSlicePlot,
    ProjectionPlot,
    SlicePlot,
    plot_2d,
)
from .profile_plotter import PhasePlot, ProfilePlot
from .streamlines import Streamlines



import yt

# Load the dataset
ds = yt.load("Enzo_64/DD0043/data0043")

# Create a volume rendering
sc = yt.create_scene(ds, field=("gas", "density"))

# Now increase the resolution
sc.camera.resolution = (1024, 1024)

# Set the camera focus to a position that is offset from the center of
# the domain
sc.camera.focus = ds.arr([0.3, 0.3, 0.3], "unitary")

# Move the camera position to the other side of the dataset
sc.camera.position = ds.arr([0, 0, 0], "unitary")

# save to disk with a custom filename and apply sigma clipping to eliminate
# very bright pixels, producing an image with better contrast.
sc.render()
sc.save("custom.png", sigma_clip=4)



import copy
from contextlib import contextmanager

from yt.fields.field_info_container import NullFunc, TranslationFunc
from yt.funcs import mylog
from yt.utilities.exceptions import YTIllDefinedFilter

# One to one mapping
filter_registry: dict[str, "ParticleFilter"] = {}


class DummyFieldInfo:
    particle_type = True
    sampling_type = "particle"


dfi = DummyFieldInfo()


class ParticleFilter:
    def __init__(self, name, function, requires, filtered_type):
        self.name = name
        self.function = function
        self.requires = requires[:]
        self.filtered_type = filtered_type

    @contextmanager
    def apply(self, dobj):
        with dobj._chunked_read(dobj._current_chunk):
            with dobj._field_type_state(self.filtered_type, dfi):
                # We won't be storing the field data from the whole read, so we
                # start by filtering now.
                filter = self.function(self, dobj)
                yield
                # Retain a reference here, and we'll filter all appropriate fields
                # later.
                fd = dobj.field_data
        for f, tr in fd.items():
            if f[0] != self.filtered_type:
                continue
            if tr.shape != filter.shape and tr.shape[0] != filter.shape[0]:
                raise YTIllDefinedFilter(self, tr.shape, filter.shape)
            else:
                d = tr[filter]
            dobj.field_data[self.name, f[1]] = d

    def available(self, field_list):
        # Note that this assumes that all the fields in field_list have the
        # same form as the 'requires' attributes.  This won't be true if the
        # fields are implicitly "all" or something.
        return all((self.filtered_type, field) in field_list for field in self.requires)

    def missing(self, field_list):
        return [
            (self.filtered_type, field)
            for field in self.requires
            if (self.filtered_type, field) not in field_list
        ]

    def wrap_func(self, field_name, old_fi):
        new_fi = copy.copy(old_fi)
        new_fi.name = (self.name, field_name[1])
        if old_fi._function == NullFunc:
            new_fi._function = TranslationFunc(old_fi.name)
        # Marking the field as inherited
        new_fi._inherited_particle_filter = True
        return new_fi


def add_particle_filter(name, function, requires=None, filtered_type="all"):
    r"""Create a new particle filter in the global namespace of filters

    A particle filter is a short name that corresponds to an algorithm for
    filtering a set of particles into a subset.  This is useful for creating new
    particle types based on a cut on a particle field, such as particle mass, ID
    or type. After defining a new filter, it still needs to be added to the
    dataset by calling
    :func:`~yt.data_objects.static_output.add_particle_filter`.

    .. note::
       Alternatively, you can make use of the
       :func:`~yt.data_objects.particle_filters.particle_filter` decorator to
       define a new particle filter.

    Parameters
    ----------
    name : string
        The name of the particle filter.  New particle fields with particle type
        set by this name will be added to any dataset that enables this particle
        filter.
    function : reference to a function
        The function that defines the particle filter.  The function should
        accept two arguments: a reference to a particle filter object and a
        reference to an abstract yt data object.  See the example below.
    requires : a list of field names
        A list of field names required by the particle filter definition.
    filtered_type : string
        The name of the particle type to be filtered.

    Examples
    --------
    >>> import yt

    >>> def _stars(pfilter, data):
    ...     return data[pfilter.filtered_type, "particle_type"] == 2

    >>> yt.add_particle_filter(
    ...     "stars", function=_stars, filtered_type="all", requires=["particle_type"]
    ... )

    >>> ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")
    >>> ds.add_particle_filter("stars")
    >>> ad = ds.all_data()
    >>> print(ad["stars", "particle_mass"])
    [  1.68243760e+38   1.65690882e+38   1.65813321e+38 ...,   2.04238266e+38
       2.04523901e+38   2.04770938e+38] g

    """
    if requires is None:
        requires = []
    filter = ParticleFilter(name, function, requires, filtered_type)
    if filter_registry.get(name, None) is not None:
        mylog.warning("The %s particle filter already exists. Overriding.", name)
    filter_registry[name] = filter


def particle_filter(name=None, requires=None, filtered_type="all"):
    r"""A decorator that adds a new particle filter

    A particle filter is a short name that corresponds to an algorithm for
    filtering a set of particles into a subset.  This is useful for creating new
    particle types based on a cut on a particle field, such as particle mass, ID
    or type.

    .. note::
       Alternatively, you can make use of the
       :func:`~yt.data_objects.particle_filters.add_particle_filter` function
       to define a new particle filter using a more declarative syntax.

    Parameters
    ----------
    name : string
        The name of the particle filter.  New particle fields with particle type
        set by this name will be added to any dataset that enables this particle
        filter.  If not set, the name will be inferred from the name of the
        filter function.
    requires : a list of field names
        A list of field names required by the particle filter definition.
    filtered_type : string
        The name of the particle type to be filtered.

    Examples
    --------
    >>> import yt

    >>> # define a filter named "stars"
    >>> @yt.particle_filter(requires=["particle_type"], filtered_type="all")
    ... def stars(pfilter, data):
    ...     return data[pfilter.filtered_type, "particle_type"] == 2

    >>> ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")
    >>> ds.add_particle_filter("stars")
    >>> ad = ds.all_data()
    >>> print(ad["stars", "particle_mass"])
    [  1.68243760e+38   1.65690882e+38   1.65813321e+38 ...,   2.04238266e+38
       2.04523901e+38   2.04770938e+38] g

    """

    def wrapper(function):
        if name is None:
            used_name = function.__name__
        else:
            used_name = name
        return add_particle_filter(used_name, function, requires, filtered_type)

    return wrapper



class TreeContainer:
    r"""A recursive data container for things like merger trees and
    clump-finder trees.

    """

    _child_attr = "children"

    def __init__(self):
        setattr(self, self._child_attr, None)

    def __iter__(self):
        yield self
        children = getattr(self, self._child_attr)
        if children is None:
            return
        for child in children:
            yield from child



from ...amrex.fields import (
    BoxlibFieldInfo,
    CastroFieldInfo,
    MaestroFieldInfo,
    NyxFieldInfo,
    WarpXFieldInfo,
)
from .._deprecation import boxlib_deprecation

boxlib_deprecation()



from numpy.testing import assert_equal

from yt.testing import fake_amr_ds, fake_particle_ds, fake_random_ds


def test_center_squeeze():
    # checks that the center is reshaped correctly

    # create and test amr, random and particle data
    check_single_ds(fake_amr_ds(fields=("Density",), units=("g/cm**3",)))
    check_single_ds(fake_random_ds(16, fields=("Density",), units=("g/cm**3",)))
    check_single_ds(fake_particle_ds(npart=100))


def check_single_ds(ds):
    # checks that the center
    center = ds.domain_center  # reference center value
    for test_shape in [(1, 3), (1, 1, 3)]:
        new_center = center.reshape(test_shape)
        assert_equal(ds.sphere(new_center, 0.25).center, center)
        assert_equal(ds.slice(0, 0.25, center=new_center).center, center)
        assert_equal(
            ds.region(new_center, [-0.25, -0.25, -0.25], [0.25, 0.25, 0.25]).center,
            center,
        )



import os
from tempfile import TemporaryDirectory

import numpy as np

from yt.frontends.ytdata.api import save_as_dataset
from yt.frontends.ytdata.data_structures import YTDataContainerDataset
from yt.loaders import load
from yt.testing import fake_amr_ds, requires_module
from yt.units import YTQuantity


@requires_module("h5py")
def test_preserve_geometric_properties():
    for geom in ("cartesian", "cylindrical", "spherical"):
        ds1 = fake_amr_ds(fields=[("gas", "density")], units=["g/cm**3"], geometry=geom)
        ad = ds1.all_data()
        with TemporaryDirectory() as tmpdir:
            tmpf = os.path.join(tmpdir, "savefile.h5")
            fn = ad.save_as_dataset(tmpf, fields=[("gas", "density")])
            ds2 = load(fn)
            assert isinstance(ds2, YTDataContainerDataset)
            dfl = ds2.derived_field_list
        assert ds1.geometry == ds2.geometry == geom

        expected = set(ds1.coordinates.axis_order)
        actual = {fname for ftype, fname in dfl}
        assert expected.difference(actual) == set()


@requires_module("h5py")
def test_default_to_cartesian():
    data = {"density": np.random.random(128)}
    ds_attrs = {"current_time": YTQuantity(10, "Myr")}
    with TemporaryDirectory() as tmpdir:
        tmpf = os.path.join(tmpdir, "savefile.h5")
        fn = save_as_dataset(ds_attrs, tmpf, data)
        ds2 = load(fn)
    assert ds2.geometry == "cartesian"



import pytest

from yt import YTQuantity
from yt.testing import fake_random_ds


def test_bad_disk_input():
    # Fixes 1768
    ds = fake_random_ds(16)

    # Test invalid 3d array
    with pytest.raises(
        TypeError,
        match=r"^Expected an array of size \(3,\), received 'list' of length 4$",
    ):
        ds.disk(ds.domain_center, [0, 0, 1, 1], (10, "kpc"), (20, "kpc"))

    # Test invalid float
    with pytest.raises(
        TypeError,
        match=(
            r"^Expected a numeric value \(or size-1 array\), "
            r"received 'unyt.array.unyt_array' of length 3$"
        ),
    ):
        ds.disk(ds.domain_center, [0, 0, 1], ds.domain_center, (20, "kpc"))

    # Test invalid float
    with pytest.raises(
        TypeError,
        match=(
            r"^Expected a numeric value \(or tuple of format \(float, String\)\), "
            r"received an inconsistent tuple '\(10, 10\)'.$"
        ),
    ):
        ds.disk(ds.domain_center, [0, 0, 1], (10, 10), (20, "kpc"))

    # Test invalid iterable
    with pytest.raises(
        TypeError,
        match=r"^Expected an iterable object, received 'unyt\.array\.unyt_quantity'$",
    ):
        ds.disk(
            ds.domain_center,
            [0, 0, 1],
            (10, "kpc"),
            (20, "kpc"),
            fields=YTQuantity(1, "kpc"),
        )

    # Test invalid object
    with pytest.raises(
        TypeError,
        match=(
            r"^Expected an object of 'yt\.data_objects\.static_output\.Dataset' type, "
            r"received 'yt\.data_objects\.selection_objects\.region\.YTRegion'$"
        ),
    ):
        ds.disk(ds.domain_center, [0, 0, 1], (10, "kpc"), (20, "kpc"), ds=ds.all_data())

    # Test valid disk
    ds.disk(ds.domain_center, [0, 0, 1], (10, "kpc"), (20, "kpc"))
    ds.disk(ds.domain_center, [0, 0, 1], 10, (20, "kpc"))



import pytest

from yt.testing import (
    fake_amr_ds,
    fake_particle_ds,
)


def test_no_print_stats():
    ds = fake_particle_ds()
    with pytest.raises(NotImplementedError):
        ds.print_stats()


def test_print_stats():
    ds = fake_amr_ds()
    ds.print_stats()



import yt

ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")  # load data

field = ("gas", "temperature")  # The field to average
weight = ("gas", "mass")  # The weight for the average

ad = ds.all_data()  # This is a region describing the entire box,
# but note it doesn't read anything in yet!

# We now use our 'quantities' call to get the average quantity
average_value = ad.quantities.weighted_average_quantity(field, weight)

print(
    "Average %s (weighted by %s) is %0.3e %s"
    % (field[1], weight[1], average_value, average_value.units)
)



from . import tests
from .data_structures import (
    YTClumpContainer,
    YTClumpTreeDataset,
    YTDataContainerDataset,
    YTGrid,
    YTGridDataset,
    YTGridHierarchy,
    YTNonspatialDataset,
    YTNonspatialGrid,
    YTNonspatialHierarchy,
    YTProfileDataset,
    YTSpatialPlotDataset,
)
from .fields import YTDataContainerFieldInfo, YTGridFieldInfo
from .io import (
    IOHandlerYTDataContainerHDF5,
    IOHandlerYTGridHDF5,
    IOHandlerYTNonspatialhdf5,
    IOHandlerYTSpatialPlotHDF5,
)
from .utilities import save_as_dataset



import cmyt  # noqa: F401
import matplotlib as mpl
import numpy as np
from matplotlib.colors import LinearSegmentedColormap

from yt.funcs import get_brewer_cmap
from yt.utilities.logger import ytLogger as mylog

from . import _colormap_data as _cm

yt_colormaps = {}


def add_colormap(name, cdict):
    """
    Adds a colormap to the colormaps available in yt for this session
    """
    # Note: this function modifies the global variable 'yt_colormaps'
    yt_colormaps[name] = LinearSegmentedColormap(name, cdict, 256)
    mpl.colormaps.register(yt_colormaps[name])


# YTEP-0040 backward compatibility layer
# yt colormaps used to be defined here, but were migrated to an external
# package, cmyt. In the process, 5 of them were renamed. We register them again here
# under their historical names to preserves backwards compatibility.

_HISTORICAL_ALIASES = {
    "arbre": "cmyt.arbre",
    "algae": "cmyt.algae",
    "bds_highcontrast": "cmyt.algae",
    "octarine": "cmyt.octarine",
    "dusk": "cmyt.dusk",
    "kamae": "cmyt.pastel",
    "kelp": "cmyt.kelp",
    "black_blueish": "cmyt.pixel_blue",
    "black_green": "cmyt.pixel_green",
    "purple_mm": "cmyt.xray",
}


def register_yt_colormaps_from_cmyt():
    """
    For backwards compatibility, register yt colormaps without the "cmyt."
    prefix, but do it in a collision-safe way.
    """

    for hist_name, alias in _HISTORICAL_ALIASES.items():
        # note that mpl.colormaps.__getitem__ returns *copies*
        cmap = mpl.colormaps[alias]
        cmap.name = hist_name
        mpl.colormaps.register(cmap)
        mpl.colormaps.register(cmap.reversed())


register_yt_colormaps_from_cmyt()

# Add colormaps in _colormap_data.py that weren't defined here
_vs = np.linspace(0, 1, 256)
for k, v in list(_cm.color_map_luts.items()):
    if k in yt_colormaps:
        continue
    cdict = {
        "red": np.transpose([_vs, v[0], v[0]]),
        "green": np.transpose([_vs, v[1], v[1]]),
        "blue": np.transpose([_vs, v[2], v[2]]),
    }
    try:
        add_colormap(k, cdict)
    except ValueError:
        # expected if another map with identical name was already registered
        mylog.warning("cannot register colormap '%s' (naming collision)", k)


def get_colormap_lut(cmap_id: tuple[str, str] | str):
    # "lut" stands for "lookup table". This function provides a consistent and
    # reusable accessor to a hidden (and by default, uninitialized) attribute
    # (`_lut`) in registered colormaps, from matplotlib or palettable.
    # colormap "lookup tables" are RGBA arrays in matplotlib,
    # and contain sufficient data to reconstruct the colormaps entirely.
    # This exists mostly for historical reasons, hence the custom output format.
    # It isn't meant as part of yt's public api.

    if isinstance(cmap_id, tuple) and len(cmap_id) == 2:
        cmap = get_brewer_cmap(cmap_id)
    elif isinstance(cmap_id, str):
        cmap = mpl.colormaps[cmap_id]
    else:
        raise TypeError(
            "Expected a string or a 2-tuple of strings as a colormap id. "
            f"Received: {cmap_id}"
        )
    if not cmap._isinit:
        cmap._init()
    r = cmap._lut[:-3, 0]
    g = cmap._lut[:-3, 1]
    b = cmap._lut[:-3, 2]
    a = np.ones(b.shape)
    return [r, g, b, a]


def show_colormaps(subset="all", filename=None):
    """
    Displays the colormaps available to yt.  Note, most functions can use
    both the matplotlib and the native yt colormaps; however, there are
    some special functions existing within image_writer.py (e.g. write_image()
    write_bitmap(), etc.), which cannot access the matplotlib
    colormaps.

    In addition to the colormaps listed, one can access the reverse of each
    colormap by appending a "_r" to any map.

    If you wish to only see certain colormaps, include them in the cmap_list
    attribute.

    Parameters
    ----------

    subset : string, or list of strings, optional

        valid values : "all", "yt_native", or list of cmap names
        default : "all"

        As mentioned above, a few functions can only access yt_native
        colormaps.  To display only the yt_native colormaps, set this
        to "yt_native".

        If you wish to only see a few colormaps side by side, you can
        include them as a list of colormap names.
        Example: ['cmyt.algae', 'gist_stern', 'cmyt.kamae', 'nipy_spectral']

    filename : string, opt

        default: None

        If filename is set, then it will save the colormaps to an output
        file.  If it is not set, it will "show" the result interactively.
    """
    from matplotlib import pyplot as plt

    a = np.outer(np.arange(0, 1, 0.01), np.ones(10))
    if subset == "all":
        maps = [
            m
            for m in plt.colormaps()
            if (not m.startswith("idl")) & (not m.endswith("_r"))
        ]
    elif subset == "yt_native":
        maps = [
            m
            for m in _cm.color_map_luts
            if (not m.startswith("idl")) & (not m.endswith("_r"))
        ]
    else:
        try:
            maps = [m for m in plt.colormaps() if m in subset]
            if len(maps) == 0:
                raise AttributeError
        except AttributeError as e:
            raise AttributeError(
                "show_colormaps requires subset attribute "
                "to be 'all', 'yt_native', or a list of "
                "valid colormap names."
            ) from e

    maps = sorted(set(maps))
    # scale the image size by the number of cmaps
    plt.figure(figsize=(2.0 * len(maps) / 10.0, 6))
    plt.subplots_adjust(top=0.7, bottom=0.05, left=0.01, right=0.99)
    l = len(maps) + 1
    for i, m in enumerate(maps):
        plt.subplot(1, l, i + 1)
        plt.axis("off")
        plt.imshow(a, aspect="auto", cmap=mpl.colormaps[m], origin="lower")
        plt.title(m, rotation=90, fontsize=10, verticalalignment="bottom")
    if filename is not None:
        plt.savefig(filename, dpi=100, facecolor="gray")
    else:
        plt.show()


def make_colormap(ctuple_list, name=None, interpolate=True):
    """
    This generates a custom colormap based on the colors and spacings you
    provide.  Enter a ctuple_list, which consists of tuples of (color, spacing)
    to return a colormap appropriate for use in yt.  If you specify a
    name, it will automatically be added to the current session as a valid
    colormap.

    Output colormap is in the format yt expects for adding a colormap to the
    current session: a dictionary with the appropriate RGB channels each
    consisting of a 256x3 array :
    First number is the number at which we are defining a color breakpoint
    Second number is the (0..1) number to interpolate to when coming *from below*
    Third number is the (0..1) number to interpolate to when coming *from above*

    Parameters
    ----------

    ctuple_list: list of (color, float) tuples
        The ctuple_list consists of pairs of (color, interval) tuples
        identifying the colors to use in the colormap and the intervals
        they take to change to the next color in the list.  A color can
        either be a string of the name of a color, or it can be an array
        of 3 floats, each representing the intensity of R, G, and B on
        a scale of 0 to 1.  Valid color names and their equivalent
        arrays are listed below.

        Any interval can be given for the different color tuples, and
        the total of all the intervals will be scaled to the 256 output
        elements.

        If a ctuple_list ends with a color and a non-zero interval,
        a white 0-interval would be added to the end to finish the
        interpolation.  To avoid finishing with white, specify your own
        zero-interval color at the end.

    name: string, optional
        If you wish this colormap to be added as a valid colormap to the
        current session, specify a name here.  Default: None

    interpolate: boolean
        Designates whether or not the colormap will interpolate between
        the colors provided or just give solid colors across the intervals.
        Default: True

    Preset Color Options
    --------------------

    'white' : np.array([255, 255, 255 ])/255.
    'gray' : np.array([130, 130, 130])/255.
    'dgray' : np.array([80, 80, 80])/255.
    'black' : np.array([0, 0, 0])/255.
    'blue' : np.array([0, 0, 255])/255.
    'dblue' : np.array([0, 0, 160])/255.
    'purple' : np.array([100, 0, 200])/255.
    'dpurple' : np.array([66, 0, 133])/255.
    'dred' : np.array([160, 0, 0])/255.
    'red' : np.array([255, 0, 0])/255.
    'orange' : np.array([255, 128, 0])/255.
    'dorange' : np.array([200,100, 0])/255.
    'yellow' : np.array([255, 255, 0])/255.
    'dyellow' : np.array([200, 200, 0])/255.
    'green' : np.array([0, 255, 0])/255.
    'dgreen' : np.array([0, 160, 0])/255.

    Examples
    --------

    To obtain a colormap that starts at black with equal intervals in green,
    blue, red, yellow in that order and interpolation between those colors.
    (In reality, it starts at black, takes an interval of 10 to interpolate to
    green, then an interval of 10 to interpolate to blue, then an interval of
    10 to interpolate to red.)

    >>> cm = make_colormap([("black", 10), ("green", 10), ("blue", 10), ("red", 0)])

    To add a colormap that has five equal blocks of solid major colors to
    the current session as "steps":

    >>> make_colormap(
    ...     [("red", 10), ("orange", 10), ("yellow", 10), ("green", 10), ("blue", 10)],
    ...     name="steps",
    ...     interpolate=False,
    ... )

    To add a colormap that looks like the French flag (i.e. equal bands of
    blue, white, and red) using your own RGB keys, then to display it:

    >>> make_colormap(
    ...     [([0, 0, 1], 10), ([1, 1, 1], 10), ([1, 0, 0], 10)],
    ...     name="french_flag",
    ...     interpolate=False,
    ... )
    >>> show_colormaps(["french_flag"])

    """
    # aliases for different colors
    color_dict = {
        "white": np.array([255, 255, 255]) / 255.0,
        "gray": np.array([130, 130, 130]) / 255.0,
        "dgray": np.array([80, 80, 80]) / 255.0,
        "black": np.array([0, 0, 0]) / 255.0,
        "blue": np.array([0, 0, 255]) / 255.0,
        "dblue": np.array([0, 0, 160]) / 255.0,
        "purple": np.array([100, 0, 200]) / 255.0,
        "dpurple": np.array([66, 0, 133]) / 255.0,
        "dred": np.array([160, 0, 0]) / 255.0,
        "red": np.array([255, 0, 0]) / 255.0,
        "orange": np.array([255, 128, 0]) / 255.0,
        "dorange": np.array([200, 100, 0]) / 255.0,
        "yellow": np.array([255, 255, 0]) / 255.0,
        "dyellow": np.array([200, 200, 0]) / 255.0,
        "green": np.array([0, 255, 0]) / 255.0,
        "dgreen": np.array([0, 160, 0]) / 255.0,
    }

    cmap = np.zeros((256, 3))

    # If the user provides a list with a non-zero final interval, it
    # doesn't make sense because you have an interval but no final
    # color to which it interpolates.  So provide a 0-length white final
    # interval to end the previous interval in white.
    if ctuple_list[-1][1] != 0:
        ctuple_list.append(("white", 0))

    # Figure out how many intervals there are total.
    rolling_index = 0
    for i, (color, interval) in enumerate(ctuple_list):
        if isinstance(color, str):
            ctuple_list[i] = (color_dict[color], interval)
        rolling_index += interval
    scale = 256.0 / rolling_index
    n = len(ctuple_list)

    # Step through each ctuple and interpolate from one color to the
    # next over the interval provided
    rolling_index = 0
    for i in range(n - 1):
        color, interval = ctuple_list[i]
        interval *= scale
        next_index = rolling_index + interval
        next_color, next_interval = ctuple_list[i + 1]

        if not interpolate:
            next_color = color

        # Interpolate the R, G, and B channels from one color to the next
        # Use np.round to make sure you're on a discrete index
        interval = int(np.round(next_index) - np.round(rolling_index))
        for j in np.arange(3):
            cmap[int(np.rint(rolling_index)) : int(np.rint(next_index)), j] = (
                np.linspace(color[j], next_color[j], num=interval)
            )

        rolling_index = next_index

    # Return a dictionary with the appropriate RGB channels each consisting of
    # a 256x3 array in the format that is expected by add_colormap() to add a
    # colormap to the session.

    # The format is as follows:
    #   First number is the number at which we are defining a color breakpoint
    #   Second number is the (0..1) number to interpolate to when coming *from below*
    #   Third number is the (0..1) number to interpolate to when coming *from above*
    _vs = np.linspace(0, 1, 256)
    cdict = {
        "red": np.transpose([_vs, cmap[:, 0], cmap[:, 0]]),
        "green": np.transpose([_vs, cmap[:, 1], cmap[:, 1]]),
        "blue": np.transpose([_vs, cmap[:, 2], cmap[:, 2]]),
    }

    if name is not None:
        add_colormap(name, cdict)

    return cdict



import yt

# Load the dataset.
ds = yt.load("GasSloshing/sloshing_nomag2_hdf5_plt_cnt_0150")

p = yt.SlicePlot(ds, "x", ("gas", "density"))

# Draw a velocity vector every 16 pixels.
p.annotate_velocity(factor=16)
p.save()



import os
import weakref
from collections import OrderedDict

import numpy as np

from yt._typing import AxisOrder
from yt.data_objects.index_subobjects.grid_patch import AMRGridPatch
from yt.data_objects.static_output import Dataset
from yt.geometry.grid_geometry_handler import GridIndex
from yt.utilities.file_handler import NetCDF4FileHandler, valid_netcdf_signature
from yt.utilities.logger import ytLogger as mylog

from .fields import CM1FieldInfo


class CM1Grid(AMRGridPatch):
    _id_offset = 0

    def __init__(self, id, index, level, dimensions):
        super().__init__(id, filename=index.index_filename, index=index)
        self.Parent = None
        self.Children = []
        self.Level = level
        self.ActiveDimensions = dimensions


class CM1Hierarchy(GridIndex):
    grid = CM1Grid

    def __init__(self, ds, dataset_type="cm1"):
        self.dataset_type = dataset_type
        self.dataset = weakref.proxy(ds)
        # for now, the index file is the dataset!
        self.index_filename = self.dataset.parameter_filename
        self.directory = os.path.dirname(self.index_filename)
        # float type for the simulation edges and must be float64 now
        self.float_type = np.float64
        super().__init__(ds, dataset_type)

    def _detect_output_fields(self):
        # build list of on-disk fields for dataset_type 'cm1'
        vnames = self.dataset.parameters["variable_names"]
        self.field_list = [("cm1", vname) for vname in vnames]

    def _count_grids(self):
        # This needs to set self.num_grids
        self.num_grids = 1

    def _parse_index(self):
        self.grid_left_edge[0][:] = self.ds.domain_left_edge[:]
        self.grid_right_edge[0][:] = self.ds.domain_right_edge[:]
        self.grid_dimensions[0][:] = self.ds.domain_dimensions[:]
        self.grid_particle_count[0][0] = 0
        self.grid_levels[0][0] = 0
        self.max_level = 0

    def _populate_grid_objects(self):
        self.grids = np.empty(self.num_grids, dtype="object")
        for i in range(self.num_grids):
            g = self.grid(i, self, self.grid_levels.flat[i], self.grid_dimensions[i])
            g._prepare_grid()
            g._setup_dx()
            self.grids[i] = g


class CM1Dataset(Dataset):
    _load_requirements = ["netCDF4"]
    _index_class = CM1Hierarchy
    _field_info_class = CM1FieldInfo

    def __init__(
        self,
        filename,
        dataset_type="cm1",
        storage_filename=None,
        units_override=None,
        unit_system="mks",
    ):
        self.fluid_types += ("cm1",)
        self._handle = NetCDF4FileHandler(filename)
        # refinement factor between a grid and its subgrid.
        self.refine_by = 1
        super().__init__(
            filename,
            dataset_type,
            units_override=units_override,
            unit_system=unit_system,
        )
        self.storage_filename = storage_filename

    def _setup_coordinate_handler(self, axis_order: AxisOrder | None) -> None:
        # ensure correct ordering of axes so plots aren't rotated (z should always be
        # on the vertical axis).
        super()._setup_coordinate_handler(axis_order)

        # type checking is deactivated in the following two lines because changing them is not
        # within the scope of the PR that _enabled_ typechecking here (#4244), but it'd be worth
        # having a careful look at *why* these warnings appear, as they may point to rotten code
        self.coordinates._x_pairs = (("x", "y"), ("y", "x"), ("z", "x"))  # type: ignore [union-attr]
        self.coordinates._y_pairs = (("x", "z"), ("y", "z"), ("z", "y"))  # type: ignore [union-attr]

    def _set_code_unit_attributes(self):
        # This is where quantities are created that represent the various
        # on-disk units.  These are the currently available quantities which
        # should be set, along with examples of how to set them to standard
        # values.
        with self._handle.open_ds() as _handle:
            length_unit = _handle.variables["xh"].units
        self.length_unit = self.quan(1.0, length_unit)
        self.mass_unit = self.quan(1.0, "kg")
        self.time_unit = self.quan(1.0, "s")
        self.velocity_unit = self.quan(1.0, "m/s")
        self.time_unit = self.quan(1.0, "s")

    def _parse_parameter_file(self):
        # This needs to set up the following items.  Note that these are all
        # assumed to be in code units; domain_left_edge and domain_right_edge
        # will be converted to YTArray automatically at a later time.
        # This includes the cosmological parameters.

        self.parameters = {}  # code-specific items
        with self._handle.open_ds() as _handle:
            # _handle here is a netcdf Dataset object, we need to parse some metadata
            # for constructing our yt ds.

            # TO DO: generalize this to be coordinate variable name agnostic in order to
            # make useful for WRF or climate data. For now, we're hard coding for CM1
            # specifically and have named the classes appropriately. Additionally, we
            # are only handling the cell-centered grid ("xh","yh","zh") at present.
            # The cell-centered grid contains scalar fields and interpolated velocities.
            dims = [_handle.dimensions[i].size for i in ["xh", "yh", "zh"]]
            xh, yh, zh = (_handle.variables[i][:] for i in ["xh", "yh", "zh"])
            self.domain_left_edge = np.array(
                [xh.min(), yh.min(), zh.min()], dtype="float64"
            )
            self.domain_right_edge = np.array(
                [xh.max(), yh.max(), zh.max()], dtype="float64"
            )

            # loop over the variable names in the netCDF file, record only those on the
            # "zh","yh","xh" grid.
            varnames = []
            for key, var in _handle.variables.items():
                if all(x in var.dimensions for x in ["time", "zh", "yh", "xh"]):
                    varnames.append(key)
            self.parameters["variable_names"] = varnames
            self.parameters["lofs_version"] = _handle.cm1_lofs_version
            self.parameters["is_uniform"] = _handle.uniform_mesh
            self.current_time = _handle.variables["time"][:][0]

            # record the dimension metadata: __handle.dimensions contains netcdf
            # objects so we need to manually copy over attributes.
            dim_info = OrderedDict()
            for dim, meta in _handle.dimensions.items():
                dim_info[dim] = meta.size
            self.parameters["dimensions"] = dim_info

        self.dimensionality = 3
        self.domain_dimensions = np.array(dims, dtype="int64")
        self._periodicity = (False, False, False)

        # Set cosmological information to zero for non-cosmological.
        self.cosmological_simulation = 0
        self.current_redshift = 0.0
        self.omega_lambda = 0.0
        self.omega_matter = 0.0
        self.hubble_constant = 0.0

    @classmethod
    def _is_valid(cls, filename: str, *args, **kwargs) -> bool:
        # This accepts a filename or a set of arguments and returns True or
        # False depending on if the file is of the type requested.
        if not valid_netcdf_signature(filename):
            return False

        if cls._missing_load_requirements():
            return False

        try:
            nc4_file = NetCDF4FileHandler(filename)
            with nc4_file.open_ds(keepweakref=True) as _handle:
                is_cm1_lofs = hasattr(_handle, "cm1_lofs_version")
                is_cm1 = hasattr(_handle, "cm1 version")  # not a typo, it is a space...

                # ensure coordinates of each variable array exists in the dataset
                coords = _handle.dimensions  # get the dataset wide coordinates
                failed_vars = []  # list of failed variables
                for var in _handle.variables:  # iterate over the variables
                    vcoords = _handle[var].dimensions  # get the dims for the variable
                    ncoords = len(vcoords)  # number of coordinates in variable
                    # number of coordinates that pass for a variable
                    coordspassed = sum(vc in coords for vc in vcoords)
                    if coordspassed != ncoords:
                        failed_vars.append(var)

                if failed_vars:
                    mylog.warning(
                        "Trying to load a cm1_lofs netcdf file but the "
                        "coordinates of the following fields do not match the "
                        "coordinates of the dataset: %s",
                        failed_vars,
                    )
                    return False

            if not is_cm1_lofs:
                if is_cm1:
                    mylog.warning(
                        "It looks like you are trying to load a cm1 netcdf file, "
                        "but at present yt only supports cm1_lofs output. Until"
                        " support is added, you can likely use"
                        " yt.load_uniform_grid() to load your cm1 file manually."
                    )
                return False
        except (OSError, AttributeError, ImportError):
            return False

        return True



import numpy as np
from numpy.testing import assert_equal

import yt
from yt.frontends.ramses.hilbert import get_cpu_list_cuboid, hilbert3d
from yt.testing import requires_file


def test_hilbert3d():
    # 8 different cases, checked against RAMSES' own implementation
    inputs = [
        [0, 0, 0],
        [1, 0, 0],
        [0, 1, 0],
        [1, 1, 0],
        [0, 0, 1],
        [1, 0, 1],
        [0, 1, 1],
        [1, 1, 1],
    ]
    outputs = [0, 1, 7, 6, 3, 2, 4, 5]

    for i, o in zip(inputs, outputs, strict=True):
        assert_equal(hilbert3d(i, 3).item(), o)


output_00080 = "output_00080/info_00080.txt"


@requires_file(output_00080)
def test_get_cpu_list():
    ds = yt.load(output_00080)

    np.random.seed(16091992)
    # These are randomly generated outputs, checked against RAMSES' own implementation
    inputs = (
        [[0.27747276, 0.30018937, 0.17916189], [0.42656026, 0.40509483, 0.29927838]],
        [[0.90660856, 0.44201328, 0.22770587], [1.09175462, 0.58017918, 0.2836648]],
        [[0.98542323, 0.58543376, 0.45858327], [1.04441105, 0.62079207, 0.58919283]],
        [[0.42274841, 0.44887745, 0.87793679], [0.52066634, 0.58936331, 1.00666222]],
        [[0.69964803, 0.65893669, 0.03660775], [0.80565696, 0.67409752, 0.11434604]],
    )
    outputs = ([0, 15], [0, 15], [0, 1, 15], [0, 13, 14, 15], [0])

    ncpu = ds.parameters["ncpu"]
    bound_keys = np.array(
        [ds.hilbert_indices[icpu][0] for icpu in range(1, ncpu + 1)]
        + [ds.hilbert_indices[ds.parameters["ncpu"]][1]],
        dtype="float64",
    )
    for i, o in zip(inputs, outputs, strict=True):
        bbox = i
        ls = list(get_cpu_list_cuboid(ds, bbox, bound_keys=bound_keys))
        assert len(ls) > 0
        assert all(np.array(o) == np.array(ls))



import yt

# load the dataset
ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")

# create our plot
p = yt.ParticlePlot(
    ds, ("all", "particle_position_x"), ("all", "particle_position_y"), color="b"
)

# zoom in a little bit
p.set_width(500, "kpc")

# save result
p.save()



from .cartesian_coordinates import CartesianCoordinateHandler
from .coordinate_handler import CoordinateHandler
from .cylindrical_coordinates import CylindricalCoordinateHandler
from .geographic_coordinates import (
    GeographicCoordinateHandler,
    InternalGeographicCoordinateHandler,
)
from .polar_coordinates import PolarCoordinateHandler
from .spec_cube_coordinates import SpectralCubeCoordinateHandler
from .spherical_coordinates import SphericalCoordinateHandler



import numpy as np

from yt.frontends.parthenon.api import ParthenonDataset
from yt.loaders import load
from yt.testing import (
    assert_allclose,
    assert_equal,
    assert_true,
    requires_file,
)
from yt.utilities.answer_testing.framework import (
    GenericArrayTest,
    data_dir_load,
    requires_ds,
    small_patch_amr,
)

_fields_parthenon_advection = (
    ("parthenon", "advected_0_0"),
    ("parthenon", "one_minus_advected"),
    ("parthenon", "one_minus_advected_sq"),
    ("parthenon", "one_minus_sqrt_one_minus_advected_sq_12"),
    ("parthenon", "one_minus_sqrt_one_minus_advected_sq_37"),
)

# Simple 2D test (advected spherical blob) with AMR from the main Parthenon test suite
# adjusted so that x1 != x2.
# Ran with `./example/advection/advection-example -i ../tst/regression/test_suites/output_hdf5/parthinput.advection parthenon/mesh/nx1=128 parthenon/mesh/x1min=-1.0 parthenon/mesh/x1max=1.0 Advection/vx=2`
# on changeset e5059ad
parthenon_advection = "parthenon_advection/advection_2d.out0.final.phdf"


@requires_ds(parthenon_advection)
def test_loading_data():
    ds = data_dir_load(parthenon_advection)
    assert_equal(str(ds), "advection_2d.out0.final")
    dd = ds.all_data()
    # test mesh dims
    vol = np.prod(ds.domain_right_edge - ds.domain_left_edge)
    assert_equal(vol, ds.quan(2.0, "code_length**3"))
    assert_allclose(dd.quantities.total_quantity("cell_volume"), vol)
    # test data
    for field in _fields_parthenon_advection:

        def field_func(name):
            return dd[name]

        yield GenericArrayTest(ds, field_func, args=[field])

    # reading data of two fields and compare against each other (data is squared in output)
    ad = ds.all_data()
    assert_allclose(
        ad[("parthenon", "one_minus_advected")] ** 2.0,
        ad[("parthenon", "one_minus_advected_sq")],
    )

    # check if the peak is in the domain center (and at the highest refinement level)
    dist_of_max_from_center = np.linalg.norm(
        ad.quantities.max_location(("parthenon", "advected_0_0"))[1:] - ds.domain_center
    )

    dx_min, dx_max = ad.quantities.extrema(("index", "dx"))
    dy_min, dy_max = ad.quantities.extrema(("index", "dy"))

    assert_true(dist_of_max_from_center < np.min((dx_min, dy_min)))


# 3D magnetized cluster center from downstream Parthenon code AthenaPK (Restart, Conserveds)
athenapk_cluster = "athenapk_cluster/athenapk_cluster.restart.00000.rhdf"

# Keplerian disk in 2D cylindrical from downstream Parthenon code AthenaPK (Data, Primitives)
athenapk_disk = "athenapk_disk/athenapk_disk.prim.00000.phdf"


@requires_file(athenapk_cluster)
def test_AthenaPK_rhdf():
    # Test that a downstream AthenaPK data set can be loaded with this Parthenon
    # frontend
    ds = data_dir_load(athenapk_cluster)
    assert isinstance(ds, ParthenonDataset)

    assert_equal(ds.domain_left_edge.in_units("code_length").v, (-0.15, -0.18, -0.2))
    assert_equal(ds.domain_right_edge.in_units("code_length").v, (0.15, 0.18, 0.2))


@requires_file(athenapk_disk)
def test_AthenaPK_phdf():
    # Test that a downstream AthenaPK data set can be loaded with this Parthenon
    # frontend
    assert isinstance(data_dir_load(athenapk_disk), ParthenonDataset)


_fields_derived = (
    ("gas", "temperature"),
    ("gas", "specific_thermal_energy"),
)

_fields_derived_cluster = (("gas", "magnetic_field_strength"),)


@requires_ds(athenapk_cluster)
def test_cluster():
    ds = data_dir_load(athenapk_cluster)
    assert_equal(str(ds), "athenapk_cluster.restart.00000")
    for test in small_patch_amr(ds, _fields_derived + _fields_derived_cluster):
        test_cluster.__name__ = test.description
        yield test


@requires_ds(athenapk_disk)
@requires_ds(athenapk_cluster)
def test_derived_fields():
    # Check that derived fields like temperature are present in downstream
    # which define them

    # Check temperature and specific thermal energy becomes defined from primitives
    ds = data_dir_load(athenapk_disk)
    dd = ds.all_data()

    for field in _fields_derived:

        def field_func(name):
            return dd[name]

        yield GenericArrayTest(ds, field_func, args=[field])

    # Check hydro, magnetic, and cooling fields defined from conserveds
    ds = data_dir_load(athenapk_cluster)
    dd = ds.all_data()

    for field in _fields_derived + _fields_derived_cluster:

        def field_func(name):
            return dd[name]

        yield GenericArrayTest(ds, field_func, args=[field])


@requires_file(athenapk_cluster)
@requires_file(athenapk_disk)
def test_adiabatic_index():
    # Read adiabiatic index from dataset parameters
    ds = data_dir_load(athenapk_cluster)
    assert_allclose(ds.gamma, 5.0 / 3.0, rtol=1e-12)

    ds = data_dir_load(athenapk_disk)
    assert_allclose(ds.gamma, 4.0 / 3.0, rtol=1e-12)

    # Change adiabatic index from dataset parameters
    ds = load(athenapk_disk, parameters={"gamma": 9.0 / 8.0})
    assert_allclose(ds.gamma, 9.0 / 8.0, rtol=1e-12)


@requires_file(athenapk_cluster)
def test_molecular_mass():
    # Read mu from dataset parameters
    ds = data_dir_load(athenapk_cluster)
    assert_allclose(float(ds.mu), 0.5925925925925926, rtol=1e-12)

    # Change He mass fraction from dataset parameters
    ds = load(athenapk_disk, parameters={"mu": 137})
    assert_equal(ds.mu, 137)


@requires_file(athenapk_cluster)
def test_units():
    # Check units in dataset are loaded correctly
    ds = data_dir_load(athenapk_cluster)
    assert_allclose(float(ds.quan(1, "code_time").in_units("Gyr")), 1, rtol=1e-12)
    assert_allclose(float(ds.quan(1, "code_length").in_units("Mpc")), 1, rtol=1e-12)
    assert_allclose(float(ds.quan(1, "code_mass").in_units("msun")), 1e14, rtol=1e-12)


@requires_file(athenapk_disk)
def test_load_cylindrical():
    # Load a cylindrical dataset of a full disk
    ds = data_dir_load(athenapk_disk)

    # Check that the domain edges match r in [0.5,2.0], theta in [0, 2pi]
    assert_equal(ds.domain_left_edge.in_units("code_length").v[:2], (0.5, 0))
    assert_equal(ds.domain_right_edge.in_units("code_length").v[:2], (2.0, 2 * np.pi))


# Sedov blast wave with curvlinear coords run with RIOT
riot_sedov_curvlinear = "riot_sedov_curvlinear/sedov.out1.final.phdf"


@requires_file(riot_sedov_curvlinear)
def test_load_riot_curvilinear():
    # Load a cylindrical dataset of a full disk
    ds = data_dir_load(riot_sedov_curvlinear)

    assert ("parthenon", "c.c.bulk.pressure") in ds.field_list

    ad = ds.all_data()
    dth = ad["index", "dtheta"]
    vol = ad["index", "cell_volume"]
    rho = ad["parthenon", "c.c.bulk.rho"]

    assert np.all(np.abs(dth - 2 * np.pi) <= 1e-12)

    total_mass = (rho * vol).sum()
    assert np.all(np.abs(total_mass.value - 0.169646) <= 1e-8)



import numpy as np


def _plot2D_root(
    seg,
    pts=None,
    txt=None,
    plotfile=None,
    point_kw=None,
    box_kw=None,
    axs=None,
    subplot_kw=None,
    gridspec_kw=None,
    fig_kw=None,
    save_kw=None,
    title=None,
    xlabel="x",
    ylabel="y",
    label_kw=None,
):
    r"""Plot a 2D kd-tree.

    Args:
        seg (list of np.ndarray): Line segments to plot defining box edges.
        pts (np.ndarray, optional): Points contained by the kdtree. Defaults to
            None if not provided and points are not plotted.
        txt (list of tuples, optional): Each tuple contains the (x, y, string)
            information for text labels to be added to the boxes. Defaults to
            None and text is not added.
        plotfile (:obj:`str`, optional): Full path to file where the plot
            should be saved. If None, the plot is displayed. Defaults to None
        point_kw (:obj:`dict`, optional): Keywords passed directly to
            :func:`matplotlib.pyplot.scatter` for drawing the points. Defaults
            to empty dict.
        box_kw (:obj:`dict`, optional): Keywords passed directly to
            :class:`matplotlib.collections.LineCollection` for drawing the
            leaf boxes. Defaults to empty dict.

        axs (:obj:`matplotlib.pyplot.Axes`, optional): Axes that should be used
            for plotting. Defaults to None and new axes are created.
        subplot_kw (:obj:`dict`, optional): Keywords passed directly to
            :meth:`matplotlib.figure.Figure.add_subplot`. Defaults to {}.
        gridspec_kw (:obj:`dict`, optional): Keywords passed directly to
            :class:`matplotlib.gridspec.GridSpec`. Defaults to empty dict.
        fig_kw (:obj:`dict`, optional): Keywords passed directly to
            :func:`matplotlib.pyplot.figure`. Defaults to empty dict.
        save_kw (:obj:`dict`, optional): Keywords passed directly to
            :func:`matplotlib.pyplot.savefig`. Defaults to empty dict.

        title (:obj:`str`, optional): Title that the plot should be given.
            Defaults to None and no title is displayed.
        xlabel (:obj:`str`, optional): Label for the x-axis. Defaults to 'x'.
        ylabel (:obj:`str`, optional): Label for the y-axis. Defaults to 'y'.
        label_kw (:obj:`dict`, optional): Keywords passed directly to
            :class:`matplotlib.text.Text` when creating box labels. Defaults
            to empty dict.

    Returns:
        :obj:`matplotlib.pyplot.Axes`: Axes containing the plot.

    """
    import matplotlib.pyplot as plt
    from matplotlib.collections import LineCollection

    if point_kw is None:
        point_kw = {}
    if box_kw is None:
        box_kw = {}
    if subplot_kw is None:
        subplot_kw = {}
    if gridspec_kw is None:
        gridspec_kw = {}
    if fig_kw is None:
        fig_kw = {}
    if save_kw is None:
        save_kw = {}
    if label_kw is None:
        label_kw = {}
    # Axes creation
    if axs is None:
        plt.close("all")
        fig, axs = plt.subplots(
            subplot_kw=subplot_kw, gridspec_kw=gridspec_kw, **fig_kw
        )

    # Labels
    if title is not None:
        axs.set_title(title)
    axs.set_xlabel(xlabel, **label_kw)
    axs.set_ylabel(ylabel, **label_kw)

    # Plot points
    if isinstance(pts, list):
        for p in pts:
            if p is not None:
                axs.scatter(p[:, 0], p[:, 1], **point_kw)
    elif pts is not None:
        axs.scatter(pts[:, 0], pts[:, 1], **point_kw)

    # Plot boxes
    lc = LineCollection(seg, **box_kw)
    axs.add_collection(lc)

    # Labels
    if txt is not None:
        # label_kw.setdefault('axes', axs)
        label_kw.setdefault("verticalalignment", "bottom")
        label_kw.setdefault("horizontalalignment", "left")
        for t in txt:
            plt.text(*t, **label_kw)

    axs.autoscale()
    axs.margins(0.1)

    # Save
    if plotfile is not None:
        plt.savefig(plotfile, **save_kw)
    else:
        plt.show()

    # Return axes
    return axs


def plot2D_serial(tree, pts=None, label_boxes=False, **kwargs):
    r"""Plot a 2D kd-tree constructed in serial.

    Parameters
    ----------

    tree: :class:`cykdtree.kdtree.PyKDTree`
        kd-tree class.
    pts: np.ndarray, optional
        Points contained by the kdtree.
    label_boxes: bool
        If True, leaves in the tree are labeled with their index. Defaults to False.

    Additional keywords are passed to :func:`cykdtree.plot._plot2D_root`.

    Returns
    -------

    :obj:`matplotlib.pyplot.Axes`: Axes containing the plot.

    """
    # Box edges
    seg = []
    for leaf in tree.leaves:
        le = leaf.left_edge
        re = leaf.right_edge
        # Top
        seg.append(np.array([[le[0], re[1]], [re[0], re[1]]], "float"))
        # Bottom
        seg.append(np.array([[le[0], le[1]], [re[0], le[1]]], "float"))
        # Left
        seg.append(np.array([[le[0], le[1]], [le[0], re[1]]], "float"))
        # Right
        seg.append(np.array([[re[0], le[1]], [re[0], re[1]]], "float"))

    # Labels
    txt = None
    if label_boxes:
        txt = []
        for leaf in tree.leaves:
            txt.append((leaf.left_edge[0], leaf.left_edge[1], "%d" % leaf.id))

    # Return axes
    return _plot2D_root(seg, pts=pts, txt=txt, **kwargs)



cdef class CutRegionSelector(SelectorObject):
    cdef set _positions
    cdef tuple _conditionals

    def __init__(self, dobj):
        axis_name = dobj.ds.coordinates.axis_name
        positions = np.array([dobj['index', axis_name[0]],
                              dobj['index', axis_name[1]],
                              dobj['index', axis_name[2]]]).T
        self._conditionals = tuple(dobj.conditionals)
        self._positions = set(tuple(position) for position in positions)

    cdef int select_bbox(self,  np.float64_t left_edge[3],
                     np.float64_t right_edge[3]) noexcept nogil:
        return 1

    cdef int select_bbox_dge(self,  np.float64_t left_edge[3],
                     np.float64_t right_edge[3]) noexcept nogil:
        return 1

    cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) noexcept nogil:
        with gil:
            if (pos[0], pos[1], pos[2]) in self._positions:
                return 1
            else:
                return 0

    cdef int select_point(self, np.float64_t pos[3]) noexcept nogil:
        return 1

    cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) noexcept nogil:
        return 1

    def _hash_vals(self):
        t = ()
        for i, c in enumerate(self._conditionals):
            t += ("conditional[%s]" % i, c)
        return ("conditionals", t)

cut_region_selector = CutRegionSelector



import numpy as np
import pytest
import unyt
from numpy.testing import assert_almost_equal, assert_equal

from yt import load_amr_grids, load_hdf5_file, load_uniform_grid
from yt.testing import _amr_grid_index, requires_file, requires_module

turb_vels = "UnigridData/turb_vels.h5"

_existing_fields = (
    "Bx",
    "By",
    "Bz",
    "Density",
    "MagneticEnergy",
    "Temperature",
    "turb_x-velocity",
    "turb_y-velocity",
    "turb_z-velocity",
    "x-velocity",
    "y-velocity",
    "z-velocity",
)


@requires_file(turb_vels)
@requires_module("h5py")
def test_load_hdf5_file():
    ds1 = load_hdf5_file(turb_vels)
    assert_equal(ds1.domain_dimensions, [256, 256, 256])
    for field_name in _existing_fields:
        assert ("stream", field_name) in ds1.field_list
    assert_equal(ds1.r[:]["ones"].size, 256 * 256 * 256)
    assert_equal(ds1.r[:]["Density"].size, 256 * 256 * 256)
    # Now we test that we get the same results regardless of our decomp
    ds2 = load_hdf5_file(turb_vels, nchunks=19)
    assert_equal(ds2.domain_dimensions, [256, 256, 256])
    assert_equal(ds2.r[:]["ones"].size, 256 * 256 * 256)
    assert_equal(ds2.r[:]["Density"].size, 256 * 256 * 256)
    assert_almost_equal(ds2.r[:]["Density"].min(), ds1.r[:]["Density"].min())
    assert_almost_equal(ds2.r[:]["Density"].max(), ds1.r[:]["Density"].max())
    assert_almost_equal(ds2.r[:]["Density"].std(), ds1.r[:]["Density"].std())
    # test that we can load this dataset with a different bounding box and length units
    ds3 = load_hdf5_file(
        turb_vels,
        bbox=np.array([[-1.0, 1.0], [-1.0, 1.0], [-1.0, 1.0]]),
        dataset_arguments={"length_unit": (1.0, "kpc")},
    )
    assert_almost_equal(ds3.domain_width, ds3.arr([2, 2, 2], "kpc"))


_x_coefficients = (100, 50, 30, 10, 20)
_y_coefficients = (20, 90, 80, 30, 30)
_z_coefficients = (50, 10, 90, 40, 40)


def _grid_data_function(grid, field_name):
    # We want N points from the cell-center to the cell-center on the other side
    x, y, z = (
        np.linspace(
            grid.LeftEdge[i] + grid.dds[i] / 2,
            grid.RightEdge[i] - grid.dds[i] / 2,
            grid.ActiveDimensions[i],
        )
        for i in (0, 1, 2)
    )
    r = np.sqrt(
        ((x.d - 0.5) ** 2)[:, None, None]
        + ((y.d - 0.5) ** 2)[None, :, None]
        + ((z.d - 0.5) ** 2)[None, None, :]
    )
    atten = np.exp(-20 * (1.1 * r**2))
    xv = sum(
        c * np.sin(2 ** (1 + i) * (x.d * np.pi * 2))
        for i, c in enumerate(_x_coefficients)
    )
    yv = sum(
        c * np.sin(2 ** (1 + i) * (y.d * np.pi * 2))
        for i, c in enumerate(_y_coefficients)
    )
    zv = sum(
        c * np.sin(2 ** (1 + i) * (z.d * np.pi * 2))
        for i, c in enumerate(_z_coefficients)
    )
    return atten * (xv[:, None, None] * yv[None, :, None] * zv[None, None, :])


def test_load_callable():
    grid_data = []
    for level, le, re, dims in _amr_grid_index:
        grid_data.append(
            {
                "level": level,
                "left_edge": le,
                "right_edge": re,
                "dimensions": dims,
                "density": _grid_data_function,
            }
        )
    ds = load_amr_grids(
        grid_data, [32, 32, 32], bbox=np.array([[0.0, 1.0], [0.0, 1.0], [0.0, 1.0]])
    )
    assert_equal(ds.r[:].sum("cell_volume"), ds.domain_width.prod())
    assert_almost_equal(ds.r[:].max("density").d, 2660218.62833899)
    assert_almost_equal(ds.r[:].min("density").d, -2660218.62833899)


def test_load_uniform_grid_callable():
    data = {"density": _grid_data_function, "my_temp": (_grid_data_function, "K")}
    ds = load_uniform_grid(
        data, [32, 32, 32], bbox=np.array([[0.0, 1.0], [0.0, 1.0], [0.0, 1.0]])
    )
    assert_equal(ds.r[:].sum("cell_volume"), ds.domain_width.prod())
    # note: the following min/max values differ from test_load_callable because
    # the grid here is coarser and the min/max values of the function are not
    # well-sampled.
    assert_almost_equal(ds.r[:].max("density").d, 1559160.37194738)
    assert_almost_equal(ds.r[:].min("density").d, -1559160.37194738)

    assert ds.r[:].min("my_temp").units == unyt.K

    with pytest.raises(RuntimeError, match="Callable functions can not be specified"):
        _ = load_uniform_grid(
            data,
            [32, 32, 32],
            bbox=np.array([[0.0, 1.0], [0.0, 1.0], [0.0, 1.0]]),
            nprocs=16,
        )




# distutils: libraries = STD_LIBS
"""
Simple readers for fortran unformatted data, specifically for the Tiger code.



"""


import numpy as np

cimport cython
cimport numpy as np
from libc.stdio cimport FILE, fclose, fopen

#cdef inline int imax(int i0, int i1):
    #if i0 > i1: return i0
    #return i1

cdef extern from "endian_swap.h":
    void FIX_SHORT( unsigned short )
    void FIX_LONG( unsigned )
    void FIX_FLOAT( float )

cdef extern from "platform_dep.h":
    void *alloca(size_t)

cdef extern from "stdio.h":
    cdef int SEEK_SET
    cdef int SEEK_CUR
    cdef int SEEK_END
    int fseek(FILE *stream, long offset, int whence)
    size_t fread(void *ptr, size_t size, size_t nmemb, FILE *stream)
    long ftell(FILE *stream)
    char *fgets(char *s, int size, FILE *stream)

@cython.boundscheck(False)
@cython.wraparound(False)
def read_and_seek(char *filename, np.int64_t offset1,
                  np.int64_t offset2, np.ndarray buffer, int bytes):
    cdef FILE *f = fopen(filename, "rb")
    cdef void *buf = <void *> buffer.data
    cdef char line[1024]
    cdef size_t n = 1023
    fseek(f, offset1, SEEK_SET)
    fgets(line, n, f)
    fseek(f, offset2, SEEK_CUR)
    fread(buf, 1, bytes, f)
    fclose(f)

def count_art_octs(char *fn, long offset,
                   int min_level, int max_level,
                   int nhydro_vars,
                   level_info):
    cdef int nchild = 8
    cdef int next_record = -1, nLevel = -1
    cdef int dummy_records[9]
    cdef int readin = -1
    cdef FILE *f = fopen(fn, "rb")
    fseek(f, offset, SEEK_SET)
    for _ in range(min_level + 1, max_level + 1):
        fread(dummy_records, sizeof(int), 2, f)
        fread(&nLevel, sizeof(int), 1, f); FIX_LONG(nLevel)
        print(level_info)
        level_info.append(nLevel)
        fread(dummy_records, sizeof(int), 2, f)
        fread(&next_record, sizeof(int), 1, f); FIX_LONG(next_record)
        print("Record size is:", next_record)
        # Offset for one record header we just read
        next_record = (nLevel * (next_record + 2*sizeof(int))) - sizeof(int)
        fseek(f, next_record, SEEK_CUR)
        # Now we skip the second section
        fread(&readin, sizeof(int), 1, f); FIX_LONG(readin)
        nhydro_vars = next_record//4-2-3 #nhvar in daniel's code
        #record length is normally 2 pad bytes, 8 + 2 hvars (the 2 is nchem)
        # and then 3 vars, but we can find nhvars only here and not in other
        # file headers
        next_record = (2*sizeof(int) + readin) * (nLevel * nchild)
        next_record -= sizeof(int)
        fseek(f, next_record, SEEK_CUR)
    print("nhvars",nhydro_vars)
    fclose(f)

def read_art_tree(char *fn, long offset,
                  int min_level, int max_level,
                  np.ndarray[np.int64_t, ndim=2] oct_indices,
                  np.ndarray[np.int64_t, ndim=1] oct_levels,
                  np.ndarray[np.int64_t, ndim=2] oct_info):
    #             np.ndarray[np.int64_t, ndim=1] oct_mask,
    #             np.ndarray[np.int64_t, ndim=1] oct_parents,

    # This accepts the filename of the ART header and an integer offset that
    # points to the start of the record *following* the reading of iOctFree and
    # nOct.  For those following along at home, we only need to read:
    #   iOctPr, iOctLv
    cdef int nchild = 8
    cdef int iOct, nLevel, ic1
    cdef np.int64_t next_record = -1
    cdef long long child_record
    cdef int iOctPs[3]
    cdef np.int64_t dummy_records[9]
    cdef int readin = -1
    cdef FILE *f = fopen(fn, "rb")
    fseek(f, offset, SEEK_SET)
    cdef int Level = -1
    cdef int * iNOLL = <int *> alloca(sizeof(int)*(max_level-min_level+1))
    cdef int * iHOLL = <int *> alloca(sizeof(int)*(max_level-min_level+1))
    cdef int iOctMax = 0
    level_offsets = [0]
    for _ in range(min_level + 1, max_level + 1):
        fread(&readin, sizeof(int), 1, f); FIX_LONG(readin)
        fread(&Level, sizeof(int), 1, f); FIX_LONG(Level)
        fread(&iNOLL[Level], sizeof(int), 1, f); FIX_LONG(iNOLL[Level])
        fread(&iHOLL[Level], sizeof(int), 1, f); FIX_LONG(iHOLL[Level])
        fread(&readin, sizeof(int), 1, f); FIX_LONG(readin)
        iOct = iHOLL[Level] - 1
        nLevel = iNOLL[Level]
        #print("Reading Hierarchy for Level", Lev, Level, nLevel, iOct)
        #print(ftell(f))
        for ic1 in range(nLevel):
            iOctMax = max(iOctMax, iOct)
            #print(readin, iOct, nLevel, sizeof(int))
            next_record = ftell(f)
            fread(&readin, sizeof(int), 1, f); FIX_LONG(readin)
            assert readin==52
            next_record += readin + sizeof(int)
            fread(iOctPs, sizeof(int), 3, f)
            FIX_LONG(iOctPs[0]); FIX_LONG(iOctPs[1]); FIX_LONG(iOctPs[2])
            oct_indices[iOct, 0] = iOctPs[0]
            oct_indices[iOct, 1] = iOctPs[1]
            oct_indices[iOct, 2] = iOctPs[2]
            oct_info[iOct, 1] = ic1
            #grid_info[iOct, 2] = iOctPr # we don't seem to need this
            fread(dummy_records, sizeof(int), 6, f) # skip Nb
            fread(&readin, sizeof(int), 1, f); FIX_LONG(readin)
            #oct_parents[iOct] = readin - 1
            fread(&readin, sizeof(int), 1, f); FIX_LONG(readin)
            oct_levels[iOct] = readin
            fread(&iOct, sizeof(int), 1, f); FIX_LONG(iOct)
            iOct -= 1
            assert next_record > 0
            fseek(f, next_record, SEEK_SET)
            fread(&readin, sizeof(int), 1, f); FIX_LONG(readin)
            assert readin==52

        level_offsets.append(ftell(f))

        #skip over the hydro variables
        #find the length of one child section
        #print('measuring child record ',)
        fread(&next_record, sizeof(int), 1, f)
        #print(next_record,)
        FIX_LONG(next_record)
        #print(next_record)
        fseek(f,ftell(f)-sizeof(int),SEEK_SET) #rewind
        #This is a sloppy fix; next_record is 64bit
        #and I don't think FIX_LONG(next_record) is working
        #correctly for 64bits
        if next_record > 4294967296L:
            next_record -= 4294967296L
        assert next_record == 56

        #find the length of all of the children section
        child_record = ftell(f) +  (next_record+2*sizeof(int))*nLevel*nchild
        #print('Skipping over hydro vars', ftell(f), child_record)
        fseek(f, child_record, SEEK_SET)

        # for ic1 in range(nLevel * nchild):
        #     fread(&next_record, sizeof(int), 1, f); FIX_LONG(next_record)
        #     fread(&idc, sizeof(int), 1, f); FIX_LONG(idc); idc -= 1 + (128**3)
        #     fread(&cm, sizeof(int), 1, f); FIX_LONG(cm)
        #     #if cm == 0: oct_mask[idc] = 1
        #     #else: total_masked += 1
        #     assert next_record > 0
        #     fseek(f, next_record - sizeof(int), SEEK_CUR)
    fclose(f)
    return level_offsets

def read_art_root_vars(char *fn, long root_grid_offset,
                    int nhydro_vars, int nx, int ny, int nz,
                    int ix, int iy, int iz, fields, var):

    cdef FILE *f = fopen(fn, "rb")
    cdef int j,l, cell_record_size = nhydro_vars * sizeof(float)
    cdef float temp = -1
    l=0
    fseek(f, root_grid_offset, SEEK_SET)
    # Now we seet out the cell we want
    cdef int my_offset = (((iz * ny) + iy) * nx + ix)
    #print(cell_record_size, my_offset, ftell(f))
    fseek(f, cell_record_size * my_offset, SEEK_CUR)
    #(((C)*GridDimension[1]+(B))*GridDimension[0]+A)
    for j in range(nhydro_vars):
        fread(&temp, sizeof(float), 1, f)
        if j in fields:
            FIX_FLOAT(temp)
            var[l]=temp
            l+=1
    fclose(f)

cdef void read_art_vars(FILE *f,
                    int min_level, int max_level, int nhydro_vars,
                    int grid_level,long grid_id,long child_offset,
                    fields,
                    np.ndarray[np.int64_t, ndim=1] level_offsets,
                    var):
    # nhydro_vars is the number of columns- 3 (adjusting for vars)
    # this is normally 10=(8+2chem species)
    cdef int record_size = 2+1+1+nhydro_vars+2
    cdef float temp = -1.0
    cdef float varpad[2]
    cdef int new_padding = -1
    cdef int padding[3]
    cdef long offset = 8*grid_id*record_size*sizeof(float)
    fseek(f, level_offsets[grid_level] + offset, SEEK_SET)
    for j in range(8): #iterate over the children
        l = 0
        fread(padding, sizeof(int), 3, f); FIX_LONG(padding[0])
        #print("Record Size", padding[0])
        # This should be replaced by an fread of nhydro_vars length
        for k in range(nhydro_vars): #iterate over the record
            fread(&temp, sizeof(float), 1, f); FIX_FLOAT(temp)
            #print(k, temp)
            if k in fields:
                var[j,l] = temp
                l += 1
        fread(varpad, sizeof(float), 2, f)
        fread(&new_padding, sizeof(int), 1, f); FIX_LONG(new_padding)
        assert(padding[0] == new_padding)

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def read_art_grid(int varindex,
              np.ndarray[np.int64_t, ndim=1] start_index,
              np.ndarray[np.int32_t, ndim=1] grid_dims,
              np.ndarray[np.float32_t, ndim=3] data,
              np.ndarray[np.uint8_t, ndim=3] filled,
              np.ndarray[np.float32_t, ndim=2] level_data,
              int level, int ref_factor,
              component_grid_info):
    cdef int gi, i, j, k, grid_id
    cdef int ir, jr, kr
    cdef int offi, offj, offk, odind
    cdef np.int64_t di, dj, dk
    cdef np.ndarray[np.int64_t, ndim=1] ogrid_info
    cdef np.ndarray[np.int64_t, ndim=1] og_start_index
    cdef np.float64_t temp_data
    cdef np.int64_t end_index[3]
    cdef int kr_offset, jr_offset, ir_offset
    cdef int to_fill = 0
    # Note that indexing into a cell is:
    #   (k*2 + j)*2 + i
    for i in range(3):
        end_index[i] = start_index[i] + grid_dims[i]
    for gi in range(len(component_grid_info)):
        ogrid_info = component_grid_info[gi]
        grid_id = ogrid_info[1]
        og_start_index = ogrid_info[3:6] #the oct left edge
        for i in range(2*ref_factor):
            di = i + og_start_index[0] * ref_factor
            if di < start_index[0] or di >= end_index[0]: continue
            ir = <int> (i / ref_factor)
            for j in range(2 * ref_factor):
                dj = j + og_start_index[1] * ref_factor
                if dj < start_index[1] or dj >= end_index[1]: continue
                jr = <int> (j / ref_factor)
                for k in range(2 * ref_factor):
                    dk = k + og_start_index[2] * ref_factor
                    if dk < start_index[2] or dk >= end_index[2]: continue
                    kr = <int> (k / ref_factor)
                    offi = di - start_index[0]
                    offj = dj - start_index[1]
                    offk = dk - start_index[2]
                    #print(offi, filled.shape[0],)
                    #print(offj, filled.shape[1],)
                    #print(offk, filled.shape[2])
                    if filled[offi, offj, offk] == 1: continue
                    if level > 0:
                        odind = (kr*2 + jr)*2 + ir
                        # Replace with an ART-specific reader
                        #temp_data = local_hydro_data.m_var_array[
                        #        level][8*offset + odind]
                        temp_data = level_data[varindex, 8*grid_id + odind]
                    else:
                        kr_offset = kr + <int> (start_index[0] / ref_factor)
                        jr_offset = jr + <int> (start_index[1] / ref_factor)
                        ir_offset = ir + <int> (start_index[2] / ref_factor)
                        odind = (kr_offset * grid_dims[0] + jr_offset)*grid_dims[1] + ir_offset
                        temp_data = level_data[varindex, odind]
                    data[offi, offj, offk] = temp_data
                    filled[offi, offj, offk] = 1
                    to_fill += 1
    return to_fill

@cython.cdivision(True)
@cython.boundscheck(True)
@cython.wraparound(False)
def fill_child_mask(np.ndarray[np.int64_t, ndim=2] file_locations,
                    np.ndarray[np.int64_t, ndim=1] grid_le,
                    np.ndarray[np.uint8_t, ndim=4] art_child_masks,
                    np.ndarray[np.uint8_t, ndim=3] child_mask):

    #loop over file_locations, for each row extracting the index & LE
    #of the oct we will pull pull from art_child_masks
    #then use the art_child_masks info to fill in child_mask
    cdef int i,ioct,x,y,z
    cdef int nocts = file_locations.shape[0]
    cdef int lex,ley,lez
    for i in range(nocts):
        ioct = file_locations[i,1] #from fortran to python indexing?
        lex = file_locations[i,3] - grid_le[0] #the oct left edge x
        ley = file_locations[i,4] - grid_le[1]
        lez = file_locations[i,5] - grid_le[2]
        for x in range(2):
            for y in range(2):
                for z in range(2):
                    child_mask[lex+x,ley+y,lez+z] = art_child_masks[ioct,x,y,z]



from collections import defaultdict

import numpy as np

from yt.funcs import mylog
from yt.utilities.io_handler import BaseParticleIOHandler
from yt.utilities.on_demand_imports import _h5py as h5py


class IOHandlerGadgetFOFHDF5(BaseParticleIOHandler):
    _dataset_type = "gadget_fof_hdf5"

    def __init__(self, ds):
        super().__init__(ds)
        self.offset_fields = set()

    def _read_fluid_selection(self, chunks, selector, fields, size):
        raise NotImplementedError(
            "IOHandlerGadgetFOFHDF5 _read_fluid_selection not implemented yet"
        )

    def _read_particle_coords(self, chunks, ptf):
        # This will read chunks and yield the results.
        for data_file in self._sorted_chunk_iterator(chunks):
            with h5py.File(data_file.filename, mode="r") as f:
                for ptype in sorted(ptf):
                    coords = data_file._get_particle_positions(ptype, f=f)
                    if coords is None:
                        continue
                    x = coords[:, 0]
                    y = coords[:, 1]
                    z = coords[:, 2]
                    yield ptype, (x, y, z), 0.0

    def _yield_coordinates(self, data_file):
        ptypes = self.ds.particle_types_raw
        with h5py.File(data_file.filename, mode="r") as f:
            for ptype in sorted(ptypes):
                pcount = data_file.total_particles[ptype]
                if pcount == 0:
                    continue
                coords = f[ptype][f"{ptype}Pos"][()].astype("float64")
                coords = np.resize(coords, (pcount, 3))
                yield ptype, coords

    def _read_offset_particle_field(self, field, data_file, fh):
        field_data = np.empty(data_file.total_particles["Group"], dtype="float64")
        fofindex = (
            np.arange(data_file.total_particles["Group"])
            + data_file.index_start["Group"]
        )
        for offset_file in data_file.offset_files:
            if fh.filename == offset_file.filename:
                ofh = fh
            else:
                ofh = h5py.File(offset_file.filename, mode="r")
            subindex = np.arange(offset_file.total_offset) + offset_file.offset_start
            substart = max(fofindex[0] - subindex[0], 0)
            subend = min(fofindex[-1] - subindex[0], subindex.size - 1)
            fofstart = substart + subindex[0] - fofindex[0]
            fofend = subend + subindex[0] - fofindex[0]
            field_data[fofstart : fofend + 1] = ofh["Subhalo"][field][
                substart : subend + 1
            ]
        return field_data

    def _read_particle_fields(self, chunks, ptf, selector):
        # Now we have all the sizes, and we can allocate
        for data_file in self._sorted_chunk_iterator(chunks):
            si, ei = data_file.start, data_file.end
            with h5py.File(data_file.filename, mode="r") as f:
                for ptype, field_list in sorted(ptf.items()):
                    pcount = data_file.total_particles[ptype]
                    if pcount == 0:
                        continue
                    coords = data_file._get_particle_positions(ptype, f=f)
                    x = coords[:, 0]
                    y = coords[:, 1]
                    z = coords[:, 2]
                    mask = selector.select_points(x, y, z, 0.0)
                    del x, y, z
                    if mask is None:
                        continue
                    for field in field_list:
                        if field in self.offset_fields:
                            field_data = self._read_offset_particle_field(
                                field, data_file, f
                            )
                        else:
                            if field == "particle_identifier":
                                field_data = (
                                    np.arange(data_file.total_particles[ptype])
                                    + data_file.index_start[ptype]
                                )
                            elif field in f[ptype]:
                                field_data = f[ptype][field][()].astype("float64")
                            else:
                                fname = field[: field.rfind("_")]
                                field_data = f[ptype][fname][()].astype("float64")
                                my_div = field_data.size / pcount
                                if my_div > 1:
                                    findex = int(field[field.rfind("_") + 1 :])
                                    field_data = field_data[:, findex]
                        data = field_data[si:ei][mask]
                        yield (ptype, field), data

    def _count_particles(self, data_file):
        si, ei = data_file.start, data_file.end
        pcount = {
            "Group": data_file.header["Ngroups_ThisFile"],
            "Subhalo": data_file.header["Nsubgroups_ThisFile"],
        }
        if None not in (si, ei):
            for ptype in pcount:
                pcount[ptype] = np.clip(pcount[ptype] - si, 0, ei - si)
        return pcount

    def _identify_fields(self, data_file):
        fields = []
        pcount = data_file.total_particles
        if sum(pcount.values()) == 0:
            return fields, {}
        with h5py.File(data_file.filename, mode="r") as f:
            for ptype in self.ds.particle_types_raw:
                if data_file.total_particles[ptype] == 0:
                    continue
                fields.append((ptype, "particle_identifier"))
                my_fields, my_offset_fields = subfind_field_list(
                    f[ptype], ptype, data_file.total_particles
                )
                fields.extend(my_fields)
                self.offset_fields = self.offset_fields.union(set(my_offset_fields))
        return fields, {}


class IOHandlerGadgetFOFHaloHDF5(IOHandlerGadgetFOFHDF5):
    _dataset_type = "gadget_fof_halo_hdf5"

    def _read_particle_coords(self, chunks, ptf):
        pass

    def _read_particle_selection(self, dobj, fields):
        rv = {}
        ind = {}
        # We first need a set of masks for each particle type
        ptf = defaultdict(list)  # ON-DISK TO READ
        fsize = defaultdict(lambda: 0)  # COUNT RV
        field_maps = defaultdict(list)  # ptypes -> fields
        unions = self.ds.particle_unions
        # What we need is a mapping from particle types to return types
        for field in fields:
            ftype, fname = field
            fsize[field] = 0
            # We should add a check for p.fparticle_unions or something here
            if ftype in unions:
                for pt in unions[ftype]:
                    ptf[pt].append(fname)
                    field_maps[pt, fname].append(field)
            else:
                ptf[ftype].append(fname)
                field_maps[field].append(field)

        # Now we allocate
        psize = {dobj.ptype: dobj.particle_number}
        for field in fields:
            if field[0] in unions:
                for pt in unions[field[0]]:
                    fsize[field] += psize.get(pt, 0)
            else:
                fsize[field] += psize.get(field[0], 0)
        for field in fields:
            if field[1] in self._vector_fields:
                shape = (fsize[field], self._vector_fields[field[1]])
            elif field[1] in self._array_fields:
                shape = (fsize[field],) + self._array_fields[field[1]]
            elif field in self.ds.scalar_field_list:
                shape = (1,)
            else:
                shape = (fsize[field],)
            rv[field] = np.empty(shape, dtype="float64")
            ind[field] = 0
        # Now we read.
        for field_r, vals in self._read_particle_fields(dobj, ptf):
            # Note that we now need to check the mappings
            for field_f in field_maps[field_r]:
                my_ind = ind[field_f]
                rv[field_f][my_ind : my_ind + vals.shape[0], ...] = vals
                ind[field_f] += vals.shape[0]
        # Now we need to truncate all our fields, since we allow for
        # over-estimating.
        for field_f in ind:
            rv[field_f] = rv[field_f][: ind[field_f]]
        return rv

    def _read_scalar_fields(self, dobj, scalar_fields):
        all_data = {}
        if not scalar_fields:
            return all_data
        pcount = 1
        with h5py.File(dobj.scalar_data_file.filename, mode="r") as f:
            for ptype, field_list in sorted(scalar_fields.items()):
                for field in field_list:
                    if field == "particle_identifier":
                        field_data = (
                            np.arange(dobj.scalar_data_file.total_particles[ptype])
                            + dobj.scalar_data_file.index_start[ptype]
                        )
                    elif field in f[ptype]:
                        field_data = f[ptype][field][()].astype("float64")
                    else:
                        fname = field[: field.rfind("_")]
                        field_data = f[ptype][fname][()].astype("float64")
                        my_div = field_data.size / pcount
                        if my_div > 1:
                            findex = int(field[field.rfind("_") + 1 :])
                            field_data = field_data[:, findex]
                    data = np.array([field_data[dobj.scalar_index]])
                    all_data[ptype, field] = data
        return all_data

    def _read_member_fields(self, dobj, member_fields):
        all_data = defaultdict(lambda: np.empty(dobj.particle_number, dtype=np.float64))
        if not member_fields:
            return all_data
        field_start = 0
        for i, data_file in enumerate(dobj.field_data_files):
            start_index = dobj.field_data_start[i]
            end_index = dobj.field_data_end[i]
            pcount = end_index - start_index
            if pcount == 0:
                continue
            field_end = field_start + end_index - start_index
            with h5py.File(data_file.filename, mode="r") as f:
                for ptype, field_list in sorted(member_fields.items()):
                    for field in field_list:
                        field_data = all_data[ptype, field]
                        if field in f["IDs"]:
                            my_data = f["IDs"][field][start_index:end_index].astype(
                                "float64"
                            )
                        else:
                            fname = field[: field.rfind("_")]
                            my_data = f["IDs"][fname][start_index:end_index].astype(
                                "float64"
                            )
                            my_div = my_data.size / pcount
                            if my_div > 1:
                                findex = int(field[field.rfind("_") + 1 :])
                                my_data = my_data[:, findex]
                        field_data[field_start:field_end] = my_data
            field_start = field_end
        return all_data

    def _read_particle_fields(self, dobj, ptf):
        # separate member particle fields from scalar fields
        scalar_fields = defaultdict(list)
        member_fields = defaultdict(list)
        for ptype, field_list in sorted(ptf.items()):
            for field in field_list:
                if (ptype, field) in self.ds.scalar_field_list:
                    scalar_fields[ptype].append(field)
                else:
                    member_fields[ptype].append(field)

        all_data = self._read_scalar_fields(dobj, scalar_fields)
        all_data.update(self._read_member_fields(dobj, member_fields))

        for field, field_data in all_data.items():
            yield field, field_data

    def _identify_fields(self, data_file):
        fields = []
        scalar_fields = []
        id_fields = {}
        with h5py.File(data_file.filename, mode="r") as f:
            for ptype in self.ds.particle_types_raw:
                fields.append((ptype, "particle_identifier"))
                scalar_fields.append((ptype, "particle_identifier"))
                my_fields, my_offset_fields = subfind_field_list(
                    f[ptype], ptype, data_file.total_particles
                )
                fields.extend(my_fields)
                scalar_fields.extend(my_fields)

                if "IDs" not in f:
                    continue
                id_fields = [(ptype, field) for field in f["IDs"]]
                fields.extend(id_fields)
        return fields, scalar_fields, id_fields, {}


def subfind_field_list(fh, ptype, pcount):
    fields = []
    offset_fields = []
    for field in fh.keys():
        if isinstance(fh[field], h5py.Group):
            my_fields, my_offset_fields = subfind_field_list(fh[field], ptype, pcount)
            fields.extend(my_fields)
            my_offset_fields.extend(offset_fields)
        else:
            if not fh[field].size % pcount[ptype]:
                my_div = fh[field].size / pcount[ptype]
                fname = fh[field].name[fh[field].name.find(ptype) + len(ptype) + 1 :]
                if my_div > 1:
                    for i in range(int(my_div)):
                        fields.append((ptype, "%s_%d" % (fname, i)))
                else:
                    fields.append((ptype, fname))
            elif (
                ptype == "Subhalo"
                and not fh[field].size % fh["/Subhalo"].attrs["Number_of_groups"]
            ):
                # These are actually Group fields, but they were written after
                # a load balancing step moved halos around and thus they do not
                # correspond to the halos stored in the Group group.
                my_div = fh[field].size / fh["/Subhalo"].attrs["Number_of_groups"]
                fname = fh[field].name[fh[field].name.find(ptype) + len(ptype) + 1 :]
                if my_div > 1:
                    for i in range(int(my_div)):
                        fields.append(("Group", "%s_%d" % (fname, i)))
                else:
                    fields.append(("Group", fname))
                offset_fields.append(fname)
            else:
                mylog.warning(
                    "Cannot add field (%s, %s) with size %d.",
                    ptype,
                    fh[field].name,
                    fh[field].size,
                )
                continue
    return fields, offset_fields



from typing import Any, Optional

import numpy as np

from yt.data_objects.selection_objects.region import YTRegion
from yt.geometry.selection_routines import (
    bbox_intersects,
    fully_contains,
)
from yt.utilities.lib.geometry_utils import get_hilbert_indices

# State diagram to compute the hilbert curve
_STATE_DIAGRAM = np.array(
    [
        [
            [1, 2, 0, 6, 11, 4, 5, 6, 10, 4, 7, 10],
            [0, 0, 0, 2, 4, 6, 4, 6, 2, 2, 4, 6],
        ],
        [
            [2, 6, 9, 0, 11, 4, 7, 1, 3, 4, 2, 3],
            [1, 7, 3, 3, 3, 5, 7, 7, 5, 1, 5, 1],
        ],
        [
            [3, 0, 10, 6, 0, 8, 5, 6, 1, 8, 11, 2],
            [3, 1, 7, 1, 5, 1, 3, 5, 3, 5, 7, 7],
        ],
        [
            [2, 7, 9, 11, 7, 8, 3, 10, 1, 8, 2, 6],
            [2, 6, 4, 0, 2, 2, 0, 4, 4, 6, 6, 0],
        ],
        [
            [4, 8, 1, 9, 5, 0, 1, 9, 10, 2, 7, 10],
            [7, 3, 1, 5, 7, 7, 5, 1, 1, 3, 3, 5],
        ],
        [
            [5, 8, 1, 0, 9, 6, 1, 4, 3, 7, 5, 3],
            [6, 4, 2, 4, 0, 4, 6, 0, 6, 0, 2, 2],
        ],
        [
            [3, 0, 11, 9, 0, 10, 11, 9, 5, 2, 8, 4],
            [4, 2, 6, 6, 6, 0, 2, 2, 0, 4, 0, 4],
        ],
        [
            [5, 7, 11, 8, 7, 6, 11, 10, 9, 3, 5, 4],
            [5, 5, 5, 7, 1, 3, 1, 3, 7, 7, 1, 3],
        ],
    ]
)


def hilbert3d(
    ijk: "np.ndarray[Any, np.dtype[np.int64]]", bit_length: int
) -> "np.ndarray[Any, np.dtype[np.float64]]":
    """Compute the order using Hilbert indexing.

    Arguments
    ---------
    ijk : (N, ndim) integer array
      The positions
    bit_length : integer
      The bit_length for the indexing.
    """
    ijk = np.atleast_2d(ijk)
    # A note here: there is a freedom in the way hilbert indices are
    # being computed (should it be xyz or yzx or zxy etc.)
    # and the yt convention is not the same as the RAMSES one.
    return get_hilbert_indices(bit_length, ijk[:, [1, 2, 0]].astype(np.int64))


def get_intersecting_cpus(
    ds,
    region: YTRegion,
    LE: Optional["np.ndarray[Any, np.dtype[np.float64]]"] = None,
    dx: float = 1.0,
    dx_cond: float | None = None,
    factor: float = 4.0,
    bound_keys: Optional["np.ndarray[Any, np.dtype[np.float64]]"] = None,
) -> set[int]:
    """
    Find the subset of CPUs that intersect the bbox in a recursive fashion.
    """
    if LE is None:
        LE = np.array([0, 0, 0], dtype="d")
    if dx_cond is None:
        bbox = region.get_bbox()
        dx_cond = float((bbox[1] - bbox[0]).min().to("code_length"))
    if bound_keys is None:
        ncpu = ds.parameters["ncpu"]
        bound_keys = np.empty(ncpu + 1, dtype="float64")
        bound_keys[:ncpu] = [ds.hilbert_indices[icpu + 1][0] for icpu in range(ncpu)]
        bound_keys[ncpu] = ds.hilbert_indices[ncpu][1]

    # If the current dx is smaller than the smallest size of the bbox
    if dx < dx_cond / factor:
        # Finish recursion
        return get_cpu_list_cuboid(ds, np.asarray([LE, LE + dx]), bound_keys)

    # If the current cell is fully within the selected region, stop recursion
    if fully_contains(region.selector, LE, dx):
        return get_cpu_list_cuboid(ds, np.asarray([LE, LE + dx]), bound_keys)

    dx /= 2

    ret = set()
    # Compute intersection of the eight subcubes with the bbox and recurse.
    for i in range(2):
        for j in range(2):
            for k in range(2):
                LE_new = LE + np.array([i, j, k], dtype="d") * dx

                if bbox_intersects(region.selector, LE_new, dx):
                    ret.update(
                        get_intersecting_cpus(
                            ds, region, LE_new, dx, dx_cond, factor, bound_keys
                        )
                    )
    return ret


def get_cpu_list_cuboid(
    ds,
    X: "np.ndarray[Any, np.dtype[np.float64]]",
    bound_keys: "np.ndarray[Any, np.dtype[np.float64]]",
) -> set[int]:
    """
    Return the list of the CPU intersecting with the cuboid containing the positions.
    Note that it will be 0-indexed.

    Parameters
    ----------
    ds : Dataset
      The dataset containing the information
    X : (N, ndim) float array
      An array containing positions. They should be between 0 and 1.
    """
    X = np.atleast_2d(X)
    if X.shape[1] != 3:
        raise NotImplementedError("This function is only implemented in 3D.")

    levelmax = ds.parameters["levelmax"]
    ndim = ds.parameters["ndim"]

    xmin, ymin, zmin = X.min(axis=0)
    xmax, ymax, zmax = X.max(axis=0)

    dmax = max(xmax - xmin, ymax - ymin, zmax - zmin)
    ilevel = int(np.ceil(-np.log2(dmax)))

    lmin = ilevel
    bit_length = lmin - 1
    maxdom = 2**bit_length

    imin, imax, jmin, jmax, kmin, kmax = 0, 0, 0, 0, 0, 0
    if bit_length > 0:
        imin = int(xmin * maxdom)
        imax = imin + 1
        jmin = int(ymin * maxdom)
        jmax = jmin + 1
        kmin = int(zmin * maxdom)
        kmax = kmin + 1

    dkey = (2 ** (levelmax + 1) / maxdom) ** ndim
    ndom = 1
    if bit_length > 0:
        ndom = 8

    ijkdom = idom, jdom, kdom = np.empty((3, 8), dtype=np.int64)

    idom[0], idom[1] = imin, imax
    idom[2], idom[3] = imin, imax
    idom[4], idom[5] = imin, imax
    idom[6], idom[7] = imin, imax

    jdom[0], jdom[1] = jmin, jmin
    jdom[2], jdom[3] = jmax, jmax
    jdom[4], jdom[5] = jmin, jmin
    jdom[6], jdom[7] = jmax, jmax

    kdom[0], kdom[1] = kmin, kmin
    kdom[2], kdom[3] = kmin, kmin
    kdom[4], kdom[5] = kmax, kmax
    kdom[6], kdom[7] = kmax, kmax

    bounding_min, bounding_max = np.zeros(ndom), np.zeros(ndom)
    if bit_length > 0:
        order_min = hilbert3d(ijkdom.T, bit_length)
    for i in range(ndom):
        if bit_length > 0:
            omin = order_min[i]
        else:
            omin = 0
        bounding_min[i] = omin * dkey
        bounding_max[i] = (omin + 1) * dkey

    cpu_min = np.searchsorted(bound_keys, bounding_min, side="right") - 1
    cpu_max = np.searchsorted(bound_keys, bounding_max, side="right")

    cpu_read: set[int] = set()

    for i in range(ndom):
        cpu_read.update(range(cpu_min[i], cpu_max[i]))

    return cpu_read
